""" Test_RSS_Command_GOCDBStatusCommand
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from datetime import datetime, timedelta

import mock
import pytest

from DIRAC import gLogger, S_OK
from DIRAC.ResourceStatusSystem.Command.DowntimeCommand import DowntimeCommand

"""
Setup
"""
gLogger.setLevel("DEBUG")

# Mock external libraries / modules not interesting for the unit test
seMock = mock.MagicMock()
seMock.options = {"SEType": "T0D1"}

mock_GOCDBClient = mock.MagicMock()
mock_RMClient = mock.MagicMock()
mock_RMClient.addOrModifyDowntimeCache.return_value = S_OK()

args = {"name": "aName", "element": "Resource", "elementType": "StorageElement"}


def test_instantiate():
    """tests that we can instantiate one object of the tested class"""

    command = DowntimeCommand()
    assert command.__class__.__name__ == "DowntimeCommand"


def test_init():
    """tests that the init method does what it should do"""

    command = DowntimeCommand()
    assert command.args == {"onlyCache": False}
    assert command.apis == {}

    command = DowntimeCommand(clients={"GOCDBClient": mock_GOCDBClient})
    assert command.args == {"onlyCache": False}
    assert command.apis == {"GOCDBClient": mock_GOCDBClient}

    command = DowntimeCommand(args)
    _args = dict(args)
    _args.update({"onlyCache": False})
    assert command.args == _args
    assert command.apis == {}


def test_doCache(mocker):
    """tests the doCache method"""
    mocker.patch("DIRAC.ResourceStatusSystem.Command.DowntimeCommand.StorageElement", return_value=seMock)
    mocker.patch(
        "DIRAC.ResourceStatusSystem.Command.DowntimeCommand.getSEHosts", return_value=S_OK(["someHost", "aSecondHost"])
    )

    command = DowntimeCommand(args, {"ResourceManagementClient": mock_RMClient})
    res = command.doCache()
    assert res["OK"] is True

    # CASE01: get ongoing DT from 2 DTs where one ongoing the other in the future
    now = datetime.utcnow()
    resFromDB = {
        "OK": True,
        "Value": (
            (
                now - timedelta(hours=2),
                "1 aRealName",
                "https://blah",
                now + timedelta(hours=3),
                "aRealName",
                now - timedelta(hours=2),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
            ),
            (
                now + timedelta(hours=12),
                "2 aRealName",
                "https://blah",
                now + timedelta(hours=14),
                "aRealName",
                now + timedelta(hours=12),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
            ),
        ),
        "Columns": [
            "StartDate",
            "DowntimeID",
            "Link",
            "EndDate",
            "Name",
            "DateEffective",
            "Description",
            "Severity",
            "LastCheckTime",
            "Element",
        ],
    }

    mock_RMClient.selectDowntimeCache.return_value = resFromDB
    command = DowntimeCommand(args, {"ResourceManagementClient": mock_RMClient})
    res = command.doCache()
    assert res["OK"] is True
    assert res["Value"]["DowntimeID"] == "1 aRealName"

    mock_RMClient.selectDowntimeCache.return_value = resFromDB
    args.update({"hours": 2})
    command = DowntimeCommand(args, {"ResourceManagementClient": mock_RMClient})
    res = command.doCache()
    assert res["OK"] is True
    assert res["Value"]["DowntimeID"] == "1 aRealName"

    # CASE02: get future DT from 2 DTs where one ongoing the other in the future
    resFromDB = {
        "OK": True,
        "Value": (
            (
                now - timedelta(hours=12),
                "1 aRealName",
                "https://blah",
                now - timedelta(hours=2),
                "aRealName",
                now - timedelta(hours=12),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
            ),
            (
                now + timedelta(hours=2),
                "2 aRealName",
                "https://blah",
                now + timedelta(hours=14),
                "aRealName",
                now + timedelta(hours=2),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
            ),
        ),
        "Columns": [
            "StartDate",
            "DowntimeID",
            "Link",
            "EndDate",
            "Name",
            "DateEffective",
            "Description",
            "Severity",
            "LastCheckTime",
            "Element",
        ],
    }

    mock_RMClient.selectDowntimeCache.return_value = resFromDB
    args.update({"hours": 3})
    command = DowntimeCommand(args, {"ResourceManagementClient": mock_RMClient})
    res = command.doCache()
    assert res["OK"] is True
    assert res["Value"]["DowntimeID"] == "2 aRealName"

    # CASE03: get DT from 2 overlapping OUTAGE DTs, one ongoing the other starting in the future
    resFromDB = {
        "OK": True,
        "Value": (
            (
                now - timedelta(hours=12),
                "1 aRealName",
                "https://blah",
                now + timedelta(hours=2),
                "aRealName",
                now - timedelta(hours=12),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
            ),
            (
                now + timedelta(hours=2),
                "2 aRealName",
                "https://blah",
                now + timedelta(hours=14),
                "aRealName",
                now + timedelta(hours=2),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
            ),
        ),
        "Columns": [
            "StartDate",
            "DowntimeID",
            "Link",
            "EndDate",
            "Name",
            "DateEffective",
            "Description",
            "Severity",
            "LastCheckTime",
            "Element",
        ],
    }

    mock_RMClient.selectDowntimeCache.return_value = resFromDB
    args.update({"hours": 0})
    command = DowntimeCommand(args, {"ResourceManagementClient": mock_RMClient})
    res = command.doCache()
    assert res["OK"] is True
    assert res["Value"]["DowntimeID"] == "1 aRealName"

    # CASE04: get DT from 2 ongoing DTs, first OUTAGE the other WARNING
    resFromDB = {
        "OK": True,
        "Value": (
            (
                now - timedelta(hours=10),
                "1 aRealName",
                "https://blah",
                now + timedelta(hours=2),
                "aRealName",
                now - timedelta(hours=12),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
            ),
            (
                now - timedelta(hours=12),
                "2 aRealName",
                "https://blah",
                now + timedelta(hours=4),
                "aRealName",
                now + timedelta(hours=2),
                "maintenance",
                "WARNING",
                now,
                "Resource",
            ),
        ),
        "Columns": [
            "StartDate",
            "DowntimeID",
            "Link",
            "EndDate",
            "Name",
            "DateEffective",
            "Description",
            "Severity",
            "LastCheckTime",
            "Element",
        ],
    }

    mock_RMClient.selectDowntimeCache.return_value = resFromDB
    args.update({"hours": 0})
    command = DowntimeCommand(args, {"ResourceManagementClient": mock_RMClient})
    res = command.doCache()
    assert res["OK"] is True
    assert res["Value"]["DowntimeID"] == "1 aRealName"

    # CASE05: get DT from 2 overlapping future DTs, the first WARNING the other OUTAGE
    resFromDB = {
        "OK": True,
        "Value": (
            (
                now + timedelta(hours=8),
                "1 aRealName",
                "https://blah",
                now + timedelta(hours=12),
                "aRealName",
                now + timedelta(hours=8),
                "maintenance",
                "WARNING",
                now,
                "Resource",
            ),
            (
                now + timedelta(hours=9),
                "2 aRealName",
                "https://blah",
                now + timedelta(hours=11),
                "aRealName",
                now + timedelta(hours=9),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
            ),
        ),
        "Columns": [
            "StartDate",
            "DowntimeID",
            "Link",
            "EndDate",
            "Name",
            "DateEffective",
            "Description",
            "Severity",
            "LastCheckTime",
            "Element",
        ],
    }

    mock_RMClient.selectDowntimeCache.return_value = resFromDB
    args.update({"hours": 10})
    command = DowntimeCommand(args, {"ResourceManagementClient": mock_RMClient})
    res = command.doCache()
    assert res["OK"] is True
    assert res["Value"]["DowntimeID"] == "2 aRealName"


@pytest.mark.parametrize(
    "downtimeCommandArgs, gocDBClientRV, expectedRes, expectedValue",
    [
        ({"element": "X"}, None, False, None),
        ({"element": "Site", "name": "aSite", "elementType": "Z"}, S_OK(), True, None),
        (
            {"element": "Resource", "name": "669 devel.edu.mk", "elementType": "Z"},
            {
                "OK": True,
                "Value": {
                    "669 devel.edu.mk": {
                        "HOSTED_BY": "MK-01-UKIM_II",
                        "DESCRIPTION": "Problem with SE server",
                        "SEVERITY": "OUTAGE",
                        "HOSTNAME": "devel.edu.mk",
                        "GOCDB_PORTAL_URL": "myURL",
                        "FORMATED_END_DATE": "2011-07-20 00:00",
                        "FORMATED_START_DATE": "2011-07-16 00:00",
                    }
                },
            },
            True,
            None,
        ),
        (
            {"element": "Resource", "name": "669 devel.edu.mk", "elementType": "Z"},
            {
                "OK": True,
                "Value": {
                    "669 devel.edu.mk": {
                        "HOSTED_BY": "MK-01-UKIM_II",
                        "DESCRIPTION": "Problem with SE server",
                        "SEVERITY": "OUTAGE",
                        "HOSTNAME": "devel.edu.mk",
                        "URL": "anotherDevel.edu.mk",
                        "GOCDB_PORTAL_URL": "myURL",
                        "FORMATED_END_DATE": "2011-07-20 00:00",
                        "FORMATED_START_DATE": "2011-07-16 00:00",
                    }
                },
            },
            True,
            None,
        ),
    ],
)
def test_doNew(downtimeCommandArgs, gocDBClientRV, expectedRes, expectedValue):
    """tests the doNew method"""

    mock_GOCDBClient.getStatus.return_value = gocDBClientRV

    command = DowntimeCommand(
        downtimeCommandArgs, {"GOCDBClient": mock_GOCDBClient, "ResourceManagementClient": mock_RMClient}
    )
    res = command.doNew()
    assert res["OK"] is expectedRes
    if res["OK"]:
        assert res["Value"] == expectedValue


def test_doMaster(mocker):
    """tests the doMaster method"""

    mocker.patch("DIRAC.ResourceStatusSystem.Command.DowntimeCommand.getGOCSites", return_value=S_OK())
    mocker.patch("DIRAC.ResourceStatusSystem.Command.DowntimeCommand.getStorageElementsHosts", return_value=S_OK())
    mocker.patch("DIRAC.ResourceStatusSystem.Command.DowntimeCommand.getFTS3Servers", return_value=S_OK())
    mocker.patch("DIRAC.ResourceStatusSystem.Command.DowntimeCommand.getGOCSites", return_value=S_OK())
    mocker.patch("DIRAC.ResourceStatusSystem.Command.DowntimeCommand.getCESiteMapping", return_value=S_OK())

    command = DowntimeCommand(
        {"element": "Resource", "name": "669 devel.edu.mk", "elementType": "Z"},
        {"GOCDBClient": mock_GOCDBClient, "ResourceManagementClient": mock_RMClient},
    )
    res = command.doMaster()
    assert res["OK"] is True



#!/usr/bin/env python
"""
Clean a given production
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("prodID: Production ID")
    _, args = Script.parseCommandLine()

    from DIRAC.ProductionSystem.Client.ProductionClient import ProductionClient

    # get arguments
    prodID = args[0]

    res = ProductionClient().setProductionStatus(prodID, "Cleaned")
    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(1)

    DIRAC.gLogger.notice("Production %s successully cleaned" % prodID)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-pilot-output
# Author :  Stuart Paterson
########################################################################
"""
Retrieve output of a Grid pilot

Example:
  $ dirac-admin-get-pilot-output https://marlb.in2p3.fr:9000/26KCLKBFtxXKHF4_ZrQjkw
  $ ls -la
  drwxr-xr-x  2 hamar marseill      2048 Feb 21 14:13 pilot_26KCLKBFtxXKHF4_ZrQjkw
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["PilotID:  Grid ID of the pilot"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    for gridID in args:

        result = diracAdmin.getPilotOutput(gridID)
        if not result["OK"]:
            errorList.append((gridID, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()



""" Collection of utilities for finding paths in the CS
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from six.moves.urllib import parse as urlparse

from DIRAC.Core.Utilities import List
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client.Helpers import Path


def getDIRACSetup():
    """Get DIRAC default setup name

    :return: str
    """
    return gConfigurationData.extractOptionFromCFG("/DIRAC/Setup")


def divideFullName(entityName, componentName=None):
    """Convert component full name to tuple

    :param str entityName: component full name, e.g.: 'Framework/ProxyManager'
    :param str componentName: component name

    :return: tuple -- contain system and component name
    """
    entityName = entityName.strip("/")
    if entityName and "/" not in entityName and componentName:
        return (entityName, componentName)
    fields = [field.strip() for field in entityName.split("/") if field.strip()]
    if len(fields) == 2:
        return tuple(fields)
    raise RuntimeError("Service (%s) name must be with the form system/service" % entityName)


def getSystemInstance(system, setup=False):
    """Find system instance name

    :param str system: system name
    :param str setup: setup name

    :return: str
    """
    optionPath = Path.cfgPath("/DIRAC/Setups", setup or getDIRACSetup(), system)
    instance = gConfigurationData.extractOptionFromCFG(optionPath)
    if not instance:
        raise RuntimeError("Option %s is not defined" % optionPath)
    return instance


def getSystemSection(system, instance=False, setup=False):
    """Get system section

    :param str system: system name
    :param str instance: instance name
    :param str setup: setup name

    :return: str -- system section path
    """
    system, _ = divideFullName(system, "_")  # for backward compatibility
    return Path.cfgPath(
        "/Systems",
        system,
        instance or getSystemInstance(system, setup=setup),
    )


def getComponentSection(system, component=False, setup=False, componentCategory="Services"):
    """Function returns the path to the component.

    :param str system: system name or component name prefixed by the system in which it is placed.
                       e.g. 'WorkloadManagement/SandboxStoreHandler'
    :param str component: component name, e.g. 'SandboxStoreHandler'
    :param str setup: Name of the setup.
    :param str componentCategory: Category of the component, it can be:
                                  'Agents', 'Services', 'Executors' or 'Databases'.

    :return: Complete path to the component
    :rtype: str

    :raise RuntimeException: If in the system - the system part does not correspond to any known system in DIRAC.

    Examples:
      getComponentSection('WorkloadManagement/SandboxStoreHandler', setup='Production', componentCategory='Services')
      getComponentSection('WorkloadManagement', 'SandboxStoreHandler', 'Production')
    """
    system, component = divideFullName(system, component)
    return Path.cfgPath(getSystemSection(system, setup=setup), componentCategory, component)


def getServiceSection(system, serviceName=False, setup=False):
    """Get service section in a system

    :param str system: system name
    :param str serviceName: service name
    :param str setup: setup name

    :return: str
    """
    return getComponentSection(system, component=serviceName, setup=setup)


def getAgentSection(system, agentName=False, setup=False):
    """Get agent section in a system

    :param str system: system name
    :param str agentName: agent name
    :param str setup: setup name

    :return: str
    """
    return getComponentSection(system, component=agentName, setup=setup, componentCategory="Agents")


def getExecutorSection(system, executorName=None, component=False, setup=False):
    """Get executor section in a system

    :param str system: system name
    :param str executorName: executor name
    :param str setup: setup name

    :return: str
    """
    return getComponentSection(system, component=executorName, setup=setup, componentCategory="Executors")


def getDatabaseSection(system, dbName=False, setup=False):
    """Get DB section in a system

    :param str system: system name
    :param str dbName: DB name
    :param str setup: setup name

    :return: str
    """
    return getComponentSection(system, component=dbName, setup=setup, componentCategory="Databases")


def getSystemURLSection(system, setup=False):
    """Get URLs section in a system

    :param str system: system name
    :param str setup: setup name

    :return: str
    """
    return Path.cfgPath(getSystemSection(system, setup=setup), "URLs")


def checkComponentURL(componentURL, system=None, component=None, pathMandatory=False):
    """Check component URL port and path. Set default ports for http scheme and raise if no port can be found.
    Set path if its mandatory or raise if its empty in this case.

    :param str componentURL: full URL, e.g.: dips://some-domain:3424/Framework/Service
    :param str system: system name
    :param str component: component name
    :param bool pathMandatory: raise error if the path could not be generated

    :return: str
    """
    url = urlparse.urlparse(componentURL)
    # Check port
    if not url.port:
        if url.scheme == "dips":
            raise RuntimeError("No port found for %s/%s URL!" % (system, component))
        url = url._replace(netloc=url.netloc + ":" + str(80 if url.scheme == "http" else 443))
    # Check path
    if not url.path.strip("/"):
        if system and component:
            url = url._replace(path="/%s/%s" % (system, component))
        elif pathMandatory:
            raise RuntimeError("No path found for %s/%s URL!" % (system, component))
    return url.geturl()


def getSystemURLs(system, setup=False, failover=False):
    """Generate url.

    :param str system: system name or full name e.g.: Framework/ProxyManager
    :param str setup: DIRAC setup name, can be defined in dirac.cfg
    :param bool failover: to add failover URLs to end of result list

    :return: dict -- complete urls. e.g. [dips://some-domain:3424/Framework/Service]
    """
    urlDict = {}
    for service in gConfigurationData.getOptionsFromCFG("%s/URLs" % getSystemSection(system, setup=setup)) or []:
        urlDict[service] = getServiceURLs(system, service, setup=setup, failover=failover)
    return urlDict


def getServiceURLs(system, service=None, setup=False, failover=False):
    """Generate url.

    :param str system: system name or full name e.g.: Framework/ProxyManager
    :param str service: service name, like 'ProxyManager'.
    :param str setup: DIRAC setup name, can be defined in dirac.cfg
    :param bool failover: to add failover URLs to end of result list

    :return: list -- complete urls. e.g. [dips://some-domain:3424/Framework/Service]
    """
    system, service = divideFullName(system, service)
    resList = []
    mainServers = None
    systemSection = getSystemSection(system, setup=setup)

    # Add failover URLs at the end of the list
    failover = "Failover" if failover else ""
    for fURLs in ["", "Failover"] if failover else [""]:
        urlList = []
        urls = List.fromChar(gConfigurationData.extractOptionFromCFG("%s/%sURLs/%s" % (systemSection, fURLs, service)))

        # Be sure that urls not None
        for url in urls or []:

            # Trying if we are refering to the list of main servers
            # which would be like dips://$MAINSERVERS$:1234/System/Component
            if "$MAINSERVERS$" in url:
                if not mainServers:
                    # Operations cannot be imported at the beginning because of a bootstrap problem
                    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

                    mainServers = Operations(setup=setup).getValue("MainServers", [])
                if not mainServers:
                    raise Exception("No Main servers defined")

                for srv in mainServers:
                    _url = checkComponentURL(url.replace("$MAINSERVERS$", srv), system, service, pathMandatory=True)
                    if _url not in urlList:
                        urlList.append(_url)
                continue

            _url = checkComponentURL(url, system, service, pathMandatory=True)
            if _url not in urlList:
                urlList.append(_url)

        # Randomize list if needed
        resList.extend(List.randomize(urlList))

    return resList


def getServiceURL(system, service=None, setup=False):
    """Generate url.

    :param str system: system name or full name e.g.: Framework/ProxyManager
    :param str service: service name, like 'ProxyManager'.
    :param str setup: DIRAC setup name, can be defined in dirac.cfg

    :return: str -- complete list of urls. e.g. dips://some-domain:3424/Framework/Service, dips://..
    """
    system, service = divideFullName(system, service)
    urls = getServiceURLs(system, service=service, setup=setup)
    return ",".join(urls) if urls else ""


def getServiceFailoverURL(system, service=None, setup=False):
    """Get failover URLs for service

    :param str system: system name or full name, like 'Framework/Service'.
    :param str service: service name, like 'ProxyManager'.
    :param str setup: DIRAC setup name, can be defined in dirac.cfg

    :return: str -- complete list of urls
    """
    system, service = divideFullName(system, service)
    systemSection = getSystemSection(system, setup=setup)
    failovers = gConfigurationData.extractOptionFromCFG("%s/FailoverURLs/%s" % (systemSection, service))
    if not failovers:
        return ""
    return ",".join([checkComponentURL(u, system, service) for u in List.fromChar(failovers, ",") if u])


def getGatewayURLs(system="", service=None):
    """Get gateway URLs for service

    :param str system: system name or full name, like 'Framework/Service'.
    :param str service: service name, like 'ProxyManager'.

    :return: list or False
    """
    if system:
        system, service = divideFullName(system, service)
    siteName = gConfigurationData.extractOptionFromCFG("/LocalSite/Site")
    if not siteName:
        return False
    gateways = gConfigurationData.extractOptionFromCFG("/DIRAC/Gateways/%s" % siteName)
    if not gateways:
        return False
    gateways = List.randomize(List.fromChar(gateways, ","))
    return [checkComponentURL(u, system, service) for u in gateways if u] if system and service else gateways



""" Class that contains client access to the JobStateUpdate handler. """

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from DIRAC.Core.Base.Client import Client, createClient


@createClient("WorkloadManagement/JobStateUpdate")
class JobStateUpdateClient(Client):
    """JobStateUpdateClient sets url for the JobStateUpdateHandler."""

    def __init__(self, url=None, **kwargs):
        """
        Sets URL for JobStateUpdate handler

        :param self: self reference
        :param url: url of the JobStateUpdateHandler
        :param kwargs: forwarded to the Base Client class
        """

        super(JobStateUpdateClient, self).__init__(**kwargs)

        if not url:
            self.serverURL = "WorkloadManagement/JobStateUpdate"

        else:
            self.serverURL = url



""" Storage Factory Class - creates instances of various Storage plugins from the Core DIRAC or extensions

    This Class has three public methods:

    getStorageName():  Resolves links in the CS to the target SE name.

    getStorage():      This creates a single storage stub based on the parameters passed in a dictionary.
                      This dictionary must have the following keys: 'StorageName','PluginName','Protocol'
                      Other optional keys are 'Port','Host','Path','SpaceToken'

    getStorages()      This takes a DIRAC SE definition and creates storage stubs for the protocols found in the CS.
                      By providing an optional list of protocols it is possible to limit the created stubs.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup

# Path to find Base SE
SE_BASE_CONFIG_PATH = "/Resources/StorageElementBases"
# Path to find concrete SE
SE_CONFIG_PATH = "/Resources/StorageElements"


class StorageFactory(object):
    def __init__(self, useProxy=False, vo=None):
        self.proxy = False
        self.proxy = useProxy
        self.resourceStatus = ResourceStatus()
        self.vo = vo
        if self.vo is None:
            result = getVOfromProxyGroup()
            if result["OK"]:
                self.vo = result["Value"]
            else:
                RuntimeError("Can not get the current VO context")
        self.remotePlugins = []
        self.localPlugins = []
        self.name = ""
        self.options = {}
        self.protocols = {}
        self.storages = []

    ###########################################################################################
    #
    # Below are public methods for obtaining storage objects
    #

    def getStorageName(self, initialName):
        return self._getConfigStorageName(initialName, "Alias")

    def getStorage(self, parameterDict, hideExceptions=False):
        """This instantiates a single storage for the details provided and doesn't check the CS."""
        # The storage name must be supplied.
        if "StorageName" in parameterDict:
            storageName = parameterDict["StorageName"]
        else:
            errStr = "StorageFactory.getStorage: StorageName must be supplied"
            gLogger.error(errStr)
            return S_ERROR(errStr)

        # PluginName must be supplied otherwise nothing with work.
        pluginName = parameterDict.get("PluginName")
        if not pluginName:
            errStr = "StorageFactory.getStorage: PluginName must be supplied"
            gLogger.error(errStr)
            return S_ERROR(errStr)

        return self.__generateStorageObject(storageName, pluginName, parameterDict, hideExceptions=hideExceptions)

    def getStorages(self, storageName, pluginList=None, hideExceptions=False):
        """Get an instance of a Storage based on the DIRAC SE name based on the CS entries CS

        :param storageName: is the DIRAC SE name i.e. 'CERN-RAW'
        :param pluginList: is an optional list of protocols if a sub-set is desired i.e ['SRM2','SRM1']

        :return: dictionary containing storage elements and information about them
        """
        self.remotePlugins = []
        self.localPlugins = []
        self.name = ""
        self.options = {}
        self.protocols = {}
        self.storages = []
        if pluginList is None:
            pluginList = []
        elif isinstance(pluginList, six.string_types):
            pluginList = [pluginList]
        if not self.vo:
            gLogger.warn("No VO information available")

        # Get the name of the storage provided
        res = self._getConfigStorageName(storageName, "Alias")
        if not res["OK"]:
            return res
        storageName = res["Value"]
        self.name = storageName

        # In case the storage is made from a base SE, get this information
        res = self._getConfigStorageName(storageName, "BaseSE")
        if not res["OK"]:
            return res
        # If the storage is derived frmo another one, keep the information
        # We initialize the seConfigPath to SE_BASE_CONFIG_PATH if there is a derivedSE, SE_CONFIG_PATH if not
        if res["Value"] != storageName:
            derivedStorageName = storageName
            storageName = res["Value"]
            seConfigPath = SE_BASE_CONFIG_PATH
        else:
            derivedStorageName = None
            seConfigPath = SE_CONFIG_PATH

        # Get the options defined in the CS for this storage
        res = self._getConfigStorageOptions(
            storageName, derivedStorageName=derivedStorageName, seConfigPath=seConfigPath
        )
        if not res["OK"]:
            return res
        self.options = res["Value"]

        # Get the protocol specific details
        res = self._getConfigStorageProtocols(
            storageName, derivedStorageName=derivedStorageName, seConfigPath=seConfigPath
        )
        if not res["OK"]:
            return res
        self.protocols = res["Value"]

        requestedLocalPlugins = []
        requestedRemotePlugins = []
        requestedProtocolDetails = []
        turlProtocols = []
        # Generate the protocol specific plug-ins
        for protocolSection, protocolDetails in self.protocols.items():
            pluginName = protocolDetails.get("PluginName", protocolSection)
            if pluginList and pluginName not in pluginList:
                continue
            protocol = protocolDetails["Protocol"]
            result = self.__generateStorageObject(
                storageName, pluginName, protocolDetails, hideExceptions=hideExceptions
            )
            if result["OK"]:
                self.storages.append(result["Value"])
                if pluginName in self.localPlugins:
                    turlProtocols.append(protocol)
                    requestedLocalPlugins.append(pluginName)
                if pluginName in self.remotePlugins:
                    requestedRemotePlugins.append(pluginName)
                requestedProtocolDetails.append(protocolDetails)
            else:
                gLogger.info(result["Message"])

        if self.storages:
            resDict = {}
            resDict["StorageName"] = self.name
            resDict["StorageOptions"] = self.options
            resDict["StorageObjects"] = self.storages
            resDict["LocalPlugins"] = requestedLocalPlugins
            resDict["RemotePlugins"] = requestedRemotePlugins
            resDict["ProtocolOptions"] = requestedProtocolDetails
            resDict["TurlProtocols"] = turlProtocols
            return S_OK(resDict)
        else:
            errStr = "StorageFactory.getStorages: Failed to instantiate any storage protocols."
            gLogger.error(errStr, self.name)
            return S_ERROR(errStr)

    ###########################################################################################
    #
    # Below are internal methods for obtaining section/option/value configuration
    #

    def _getConfigStorageName(self, storageName, referenceType, seConfigPath=SE_CONFIG_PATH):
        """
        This gets the name of the storage the configuration service.
        If the storage is a reference to another SE the resolution is performed.

        :param storageName: is the storage section to check in the CS
        :param referenceType: corresponds to an option inside the storage section
        :param seConfigPath: the path of the storage section.
                                It can be /Resources/StorageElements or StorageElementBases

        :return: the name of the storage
        """
        configPath = "%s/%s" % (seConfigPath, storageName)
        res = gConfig.getOptions(configPath)
        if not res["OK"]:
            errStr = "StorageFactory._getConfigStorageName: Failed to get storage options"
            gLogger.error(errStr, res["Message"])
            return S_ERROR(errStr)
        if not res["Value"]:
            errStr = "StorageFactory._getConfigStorageName: Supplied storage doesn't exist."
            gLogger.error(errStr, configPath)
            return S_ERROR(errStr)
        if referenceType in res["Value"]:
            configPath = cfgPath(seConfigPath, storageName, referenceType)
            referenceName = gConfig.getValue(configPath)
            # We first look into the BaseStorageElements section.
            # If not, we look into the StorageElements section
            # (contrary to BaseSE, it's OK for an Alias to be in the StorageElements section)
            result = self._getConfigStorageName(referenceName, "Alias", seConfigPath=SE_BASE_CONFIG_PATH)
            if not result["OK"]:
                # Since it is not in the StorageElementBases section, check in the StorageElements section
                result = self._getConfigStorageName(referenceName, "Alias", seConfigPath=SE_CONFIG_PATH)
                if not result["OK"]:
                    return result
            resolvedName = result["Value"]
        else:
            resolvedName = storageName
        return S_OK(resolvedName)

    def _getConfigStorageOptions(self, storageName, derivedStorageName=None, seConfigPath=SE_CONFIG_PATH):
        """
        Get the options associated to the StorageElement as defined in the CS

        :param storageName: is the storage section to check in the CS
        :param seConfigPath: the path of the storage section.
                                It can be /Resources/StorageElements or StorageElementBases
        :param derivedStorageName: is the storage section of a derived storage if it inherits from a base

        :return: options associated to the StorageElement as defined in the CS
        """
        optionsDict = {}

        # We first get the options of the baseSE, and then overwrite with the derivedSE
        for seName in (storageName, derivedStorageName) if derivedStorageName else (storageName,):
            storageConfigPath = cfgPath(seConfigPath, seName)
            res = gConfig.getOptions(storageConfigPath)
            if not res["OK"]:
                errStr = "StorageFactory._getStorageOptions: Failed to get storage options."
                gLogger.error(errStr, "%s: %s" % (seName, res["Message"]))
                return S_ERROR(errStr)
            for option in set(res["Value"]) - set(("ReadAccess", "WriteAccess", "CheckAccess", "RemoveAccess")):
                optionConfigPath = cfgPath(storageConfigPath, option)
                default = [] if option in ["VO", "AccessProtocols", "WriteProtocols"] else ""
                optionsDict[option] = gConfig.getValue(optionConfigPath, default)
            # We update the seConfigPath in order to find option in derivedSE now
            seConfigPath = SE_CONFIG_PATH

        # The status is that of the derived SE only
        seName = derivedStorageName if derivedStorageName else storageName
        res = self.resourceStatus.getElementStatus(seName, "StorageElement")
        if not res["OK"]:
            errStr = "StorageFactory._getStorageOptions: Failed to get storage status"
            gLogger.error(errStr, "%s: %s" % (seName, res["Message"]))
            return S_ERROR(errStr)

        # For safety, we did not add the ${statusType}Access keys
        # this requires modifications in the StorageElement class

        # We add the dictionary with the statusTypes and values
        # { 'statusType1' : 'status1', 'statusType2' : 'status2' ... }
        optionsDict.update(res["Value"][seName])

        return S_OK(optionsDict)

    def __getProtocolsSections(self, storageName, seConfigPath=SE_CONFIG_PATH):
        """
        Get the protocols of a specific storage section

        :param storageName: is the storage section to check in the CS
        :param seConfigPath: the path of the storage section.
                                It can be /Resources/StorageElements or StorageElementBases

        :return: list of protocol section names
        """
        storageConfigPath = cfgPath(seConfigPath, storageName)
        res = gConfig.getSections(storageConfigPath)
        if not res["OK"]:
            errStr = "StorageFactory._getConfigStorageProtocols: Failed to get storage sections"
            gLogger.error(errStr, "%s: %s" % (storageName, res["Message"]))
            return S_ERROR(errStr)
        protocolSections = res["Value"]
        return S_OK(protocolSections)

    def _getConfigStorageProtocols(self, storageName, derivedStorageName=None, seConfigPath=SE_CONFIG_PATH):
        """
        Make a dictionary of protocols with the information associated. Merge with a base SE if it exists

        :param storageName: is the storage section to check in the CS
        :param seConfigPath: the path of the storage section.
                                It can be /Resources/StorageElements or StorageElementBases
        :param derivedStorageName: is the storage section of a derived storage if it inherits from a base

        :return: dictionary of protocols like {protocolSection: {protocolOptions}}
        """
        # Get the sections
        res = self.__getProtocolsSections(storageName, seConfigPath=seConfigPath)
        if not res["OK"]:
            return res
        protocolSections = res["Value"]
        sortedProtocolSections = sorted(protocolSections)

        # Get the details for each section in a dictionary
        for protocolSection in sortedProtocolSections:
            res = self._getConfigStorageProtocolDetails(storageName, protocolSection, seConfigPath=seConfigPath)
            if not res["OK"]:
                return res
            self.protocols[protocolSection] = res["Value"]
        if derivedStorageName:
            # We may have parameters overwriting the baseSE protocols
            res = self.__getProtocolsSections(derivedStorageName, seConfigPath=SE_CONFIG_PATH)
            if not res["OK"]:
                return res
            for protocolSection in res["Value"]:
                res = self._getConfigStorageProtocolDetails(
                    derivedStorageName, protocolSection, seConfigPath=SE_CONFIG_PATH, checkAccess=False
                )
                if not res["OK"]:
                    return res
                detail = res["Value"]
                # If we found the plugin section from which we inherit
                inheritanceMatched = False
                for baseStorageProtocolSection in protocolSections:
                    if protocolSection == baseStorageProtocolSection:
                        inheritanceMatched = True
                        for key, val in detail.items():
                            if val:
                                self.protocols[protocolSection][key] = val
                        break
                # If not matched, consider it a new protocol
                if not inheritanceMatched:
                    self.protocols[protocolSection] = detail
        return S_OK(self.protocols)

    def _getConfigStorageProtocolDetails(
        self, storageName, protocolSection, seConfigPath=SE_CONFIG_PATH, checkAccess=True
    ):
        """
          Parse the contents of the protocol block

        :param storageName: is the storage section to check in the CS
        :param protocolSection: name of the protocol section to find information
        :param seConfigPath: the path of the storage section.
                                  It can be /Resources/StorageElements or StorageElementBases
        :param checkAccess: if not set, don't complain if "Access" is not in the section

        :return: dictionary of the protocol options
        """
        # First obtain the options that are available
        protocolConfigPath = cfgPath(seConfigPath, storageName, protocolSection)
        res = gConfig.getOptions(protocolConfigPath)
        if not res["OK"]:
            errStr = "StorageFactory.__getProtocolDetails: Failed to get protocol options."
            gLogger.error(errStr, "%s: %s" % (storageName, protocolSection))
            return S_ERROR(errStr)
        options = res["Value"]

        # We must have certain values internally even if not supplied in CS
        protocolDict = {"Access": "", "Host": "", "Path": "", "Port": "", "Protocol": "", "SpaceToken": "", "WSUrl": ""}
        for option in options:
            configPath = cfgPath(protocolConfigPath, option)
            optionValue = gConfig.getValue(configPath, "")
            protocolDict[option] = optionValue

        # Evaluate the base path taking into account possible VO specific setting
        if self.vo:
            result = gConfig.getOptionsDict(cfgPath(protocolConfigPath, "VOPath"))
            voPath = ""
            if result["OK"]:
                voPath = result["Value"].get(self.vo, "")
            if voPath:
                protocolDict["Path"] = voPath

        # Now update the local and remote protocol lists.
        # A warning will be given if the Access option is not set and the plugin is not already in remote or local.
        plugin = protocolDict.get("PluginName", protocolSection)
        if protocolDict["Access"].lower() == "remote":
            self.remotePlugins.append(plugin)
        elif protocolDict["Access"].lower() == "local":
            self.localPlugins.append(plugin)
        # If it is a derived SE, this is normal, no warning
        elif checkAccess and protocolSection not in self.protocols:
            errStr = (
                "StorageFactory.__getProtocolDetails: The 'Access' option \
      for %s:%s is neither 'local' or 'remote'."
                % (storageName, protocolSection)
            )
            gLogger.warn(errStr)

        return S_OK(protocolDict)

    ###########################################################################################
    #
    # Below is the method for obtaining the object instantiated for a provided storage configuration
    #

    def __generateStorageObject(self, storageName, pluginName, parameters, hideExceptions=False):
        """
        Generate a Storage Element from parameters collected

        :param storageName: is the storage section to check in the CS
        :param pluginName: name of the plugin used. Example: GFAL2_XROOT, GFAL2_SRM2...
        :param parameters: dictionary of protocol details.
        """

        storageType = pluginName
        if self.proxy:
            storageType = "Proxy"

        objectLoader = ObjectLoader()
        result = objectLoader.loadObject("Resources.Storage.%sStorage" % storageType, hideExceptions=hideExceptions)
        if not result["OK"]:
            gLogger.error("Failed to load storage object: %s" % result["Message"])
            return result

        storageClass = result["Value"]
        try:
            storage = storageClass(storageName, parameters)
        except Exception as x:
            errStr = "StorageFactory._generateStorageObject: Failed to instantiate %s: %s" % (storageName, x)
            gLogger.exception(errStr)
            return S_ERROR(errStr)

        return S_OK(storage)



""" TaskManager contains TaskBase, which is inherited by WorkflowTasks and RequestTasks modules,
for managing jobs and requests tasks
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# pylint: disable=protected-access


import time

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.TransformationSystem.Agent.TransformationAgentsUtilities import TransformationAgentsUtilities

COMPONENT_NAME = "TaskManager"


class TaskBase(TransformationAgentsUtilities):
    """The other classes inside here inherits from this one."""

    def __init__(self, transClient=None, logger=None):
        """c'tor"""

        if not transClient:
            self.transClient = TransformationClient()
        else:
            self.transClient = transClient

        if not logger:
            self.log = gLogger.getSubLogger("TaskBase")
        else:
            self.log = logger

        self.pluginLocation = "DIRAC.TransformationSystem.Client.TaskManagerPlugin"

        super(TaskBase, self).__init__()

    def prepareTransformationTasks(
        self,
        _transBody,
        _taskDict,  # pylint: disable=no-self-use, unused-argument
        owner="",
        ownerGroup="",
        ownerDN="",
        bulkSubmissionFlag=False,
    ):  # pylint: disable=unused-argument
        """To make sure the method is implemented in the derived class"""
        if owner or ownerGroup or ownerDN or bulkSubmissionFlag:  # Makes pylint happy
            pass
        return S_ERROR("Not implemented")

    def submitTransformationTasks(self, _taskDict):  # pylint: disable=no-self-use
        """To make sure the method is implemented in the derived class"""
        return S_ERROR("Not implemented")

    def submitTasksToExternal(self, _task):  # pylint: disable=no-self-use
        """To make sure the method is implemented in the derived class"""
        return S_ERROR("Not implemented")

    def updateDBAfterTaskSubmission(self, taskDict):
        """Sets tasks status after the submission to "Submitted", in case of success"""
        updated = 0
        startTime = time.time()
        for taskID, task in taskDict.items():
            transID = task["TransformationID"]
            if task["Success"]:
                res = self.transClient.setTaskStatusAndWmsID(transID, taskID, "Submitted", str(task["ExternalID"]))
                if not res["OK"]:
                    self._logWarn(
                        "Failed to update task status after submission",
                        "%s %s" % (task["ExternalID"], res["Message"]),
                        transID=transID,
                        method="updateDBAfterSubmission",
                    )
                updated += 1
        if updated:
            self._logInfo(
                "Updated %d tasks in %.1f seconds" % (updated, time.time() - startTime),
                transID=transID,
                method="updateDBAfterSubmission",
            )
        return S_OK()

    def updateTransformationReservedTasks(self, _taskDicts):  # pylint: disable=no-self-use
        """To make sure the method is implemented in the derived class"""
        return S_ERROR("Not implemented")

    def getSubmittedTaskStatus(self, _taskDicts):  # pylint: disable=no-self-use
        """To make sure the method is implemented in the derived class"""
        return S_ERROR("Not implemented")

    def getSubmittedFileStatus(self, _fileDicts):  # pylint: disable=no-self-use
        """To make sure the method is implemented in the derived class"""
        return S_ERROR("Not implemented")



"""
   DIRAC - Distributed Infrastructure with Remote Agent Control

   The distributed data production and analysis system of LHCb and other VOs.

   DIRAC is a software framework for distributed computing which
   allows to integrate various computing resources in a single
   system. At the same time it integrates all kinds of computing
   activities like Monte Carlo simulations, data processing, or
   final user analysis.

   It is build as number of cooperating systems:
    - Accounting
    - Configuration
    - Core
      - Base
      - DISET
      - Security
      - Utilities
      - Workflow
    - Framework
    - RequestManagement
    - Resources
    - Transformation

    Which are used by other system providing functionality to
    the end user:
    - DataManagement
    - Interfaces
    - ResourceStatus
    - StorageManagement
    - WorkloadManagement

    It defines the following data members:
    - majorVersion:  DIRAC Major version number
    - minorVersion:  DIRAC Minor version number
    - patchLevel:    DIRAC Patch level number
    - preVersion:    DIRAC Pre release number
    - version:       DIRAC version string

    - errorMail:     mail address for important errors
    - alarmMail:     mail address for important alarms

    - pythonPath:    absolute real path to the directory that contains this file
    - rootPath:      absolute real path to the parent of DIRAC.pythonPath

    It loads Modules from :
    - DIRAC.Core.Utililies

    It loads:
    - S_OK:           OK return structure
    - S_ERROR:        ERROR return structure
    - gLogger:        global Logger object
    - gConfig:        global Config object

    It defines the following functions:
    - abort:          aborts execution
    - exit:           finish execution using callbacks
    - siteName:       returns DIRAC name for current site

    - getPlatform():      DIRAC platform string for current host
    - getPlatformTuple(): DIRAC platform tuple for current host

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
import os
import platform as pyPlatform
from pkgutil import extend_path

import re
import six

__path__ = extend_path(__path__, __name__)

# Set the environment variable such that openssl accepts proxy cert
# Sadly, this trick was removed in openssl >= 1.1.0
# https://github.com/openssl/openssl/commit/8e21938ce3a5306df753eb40a20fe30d17cf4a68
# Lets see if they would accept to put it back
# https://github.com/openssl/openssl/issues/8177
os.environ["OPENSSL_ALLOW_PROXY_CERTS"] = "True"

__RCSID__ = "$Id$"

# Now that's one hell of a hack :)
# _strptime is not thread safe, resulting in obscure callstack
# whenever you would have multiple threads and calling datetime.datetime.strptime
# (AttributeError: 'module' object has no attribute '_strptime')
# Importing _strptime before instantiating the threads seem to be a working workaround
import _strptime

# Define Version
if six.PY3:
    from pkg_resources import get_distribution, DistributionNotFound

    try:
        __version__ = get_distribution(__name__).version
        version = __version__
    except DistributionNotFound:
        # package is not installed
        version = "Unknown"
else:
    majorVersion = 7
    minorVersion = 3
    patchLevel = 12
    preVersion = 0

    version = "v%sr%s" % (majorVersion, minorVersion)
    # Make it so that __version__ is always PEP-440 style
    __version__ = "%s.%s" % (majorVersion, minorVersion)
    if patchLevel:
        version = "%sp%s" % (version, patchLevel)
        __version__ += ".%s" % patchLevel
    if preVersion:
        version = "%s-pre%s" % (version, preVersion)
        __version__ += "a%s" % preVersion

errorMail = "dirac.alarms@gmail.com"
alarmMail = "dirac.alarms@gmail.com"


def isPy3VersionNumber(releaseVersion):
    """Returns True if the releaseVersion is a PEP-440 style string.
    This is the `is_canonical` function defined in PEP-440 Appendix B

    :param str releaseVersion: The software version to use
    """
    return (
        re.match(
            r"^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$",
            releaseVersion,
        )
        is not None
    )


def convertToPy3VersionNumber(releaseVersion):
    """Convert the releaseVersion into a PEP-440 style string

    :param str releaseVersion: The software version to use
    """
    VERSION_PATTERN = re.compile(r"^(?:v)?(\d+)[r\.](\d+)(?:[p\.](\d+))?(?:(?:-pre|a)?(\d+))?$")

    match = VERSION_PATTERN.match(releaseVersion)
    # If the regex fails just return the original version
    if not match:
        return releaseVersion
    major, minor, patch, pre = match.groups()
    version = major + "." + minor
    version += "." + (patch or "0")
    if pre:
        version += "a" + pre
    return version


def _computeRootPath(rootPath):
    """Compute the root of the DIRAC installation

    Detects if the installation is a server-style versioned installation by
    recognising a folder structure like: ``versions/vX.Y.Z-$(uname -m)-TIMESTAMP/``

    :param str rootPath: The result of ``sys.base_prefix``
    :return: bool
    """
    import re
    from pathlib import Path  # pylint: disable=import-error

    rootPath = Path(rootPath).resolve()
    versionsPath = rootPath.parent
    if versionsPath.parent.name != "versions":
        return str(rootPath)
    # VERSION-INSTALL_TIME
    pattern1 = re.compile(r"v(\d+\.\d+\.\d+[^\-]*)\-(\d+)")
    # $(uname -s)-$(uname -m)
    pattern2 = re.compile(r"([^\-]+)-([^\-]+)")
    if pattern1.fullmatch(versionsPath.name) and pattern2.fullmatch(rootPath.name):
        # This is a versioned install
        return str(versionsPath.parent.parent)
    else:
        return str(rootPath)


# Set rootPath of DIRAC installation
if six.PY3:
    rootPath = _computeRootPath(sys.base_prefix)  # pylint: disable=no-member
else:
    pythonPath = os.path.realpath(__path__[0])
    rootPath = os.path.dirname(pythonPath)

# Import DIRAC.Core.Utils modules

# from DIRAC.Core.Utilities import *
from DIRAC.Core.Utilities.Network import getFQDN

from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR


# Logger
from DIRAC.FrameworkSystem.Client.Logger import gLogger

# Configuration client
from DIRAC.ConfigurationSystem.Client.Config import gConfig


__siteName = False


def siteName():
    """
    Determine and return DIRAC name for current site
    """
    global __siteName
    if not __siteName:
        __siteName = gConfig.getValue("/LocalSite/Site")
        if not __siteName:
            # Some Defaults if not present in the configuration
            FQDN = getFQDN()
            if len(FQDN.split(".")) > 2:
                # Use the last component of the FQDN as country code if there are more than 2 components
                __siteName = "DIRAC.Client.%s" % FQDN.split(".")[-1]
            else:
                # else use local as country code
                __siteName = "DIRAC.Client.local"

    return __siteName


# platform detection
from DIRAC.Core.Utilities.Platform import getPlatformString, getPlatform, getPlatformTuple


def exit(exitCode=0):
    """
    Finish execution using callbacks
    """
    sys.exit(exitCode)


def abort(exitCode, *args, **kwargs):
    """
    Abort execution
    """
    try:
        gLogger.fatal(*args, **kwargs)
        os._exit(exitCode)
    except OSError:
        gLogger.exception("Error while executing DIRAC.abort")
        os._exit(exitCode)


def extension_metadata():
    return {
        "primary_extension": True,
        "priority": 0,
        "setups": {
            "DIRAC-Certification": "https://lbcertifdirac70.cern.ch:9135/Configuration/Server",
            "DIRAC-CertifOauth": "dips://lbcertifdiracoauth.cern.ch:9135/Configuration/Server",
        },
    }



""" Class that contains client access to the StorageManagerDB handler.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import random
import errno

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities.DErrno import cmpError
from DIRAC.Core.Utilities.Proxy import UserProxy
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.Resources.Storage.StorageElement import StorageElement


def getFilesToStage(lfnList, jobState=None, checkOnlyTapeSEs=None, jobLog=None):
    """Utility that returns out of a list of LFNs those files that are offline,
    and those for which at least one copy is online
    """
    if not lfnList:
        return S_OK({"onlineLFNs": [], "offlineLFNs": {}, "failedLFNs": [], "absentLFNs": {}})

    dm = DataManager()
    if isinstance(lfnList, six.string_types):
        lfnList = [lfnList]

    lfnListReplicas = dm.getReplicasForJobs(lfnList, getUrl=False)
    if not lfnListReplicas["OK"]:
        return lfnListReplicas

    offlineLFNsDict = {}
    onlineLFNs = {}
    offlineLFNs = {}
    absentLFNs = {}
    failedLFNs = set()
    if lfnListReplicas["Value"]["Failed"]:
        # Check if files are not existing
        for lfn, reason in lfnListReplicas["Value"]["Failed"].items():
            # FIXME: awful check until FC returns a proper error
            if cmpError(reason, errno.ENOENT) or "No such file" in reason:
                # The file doesn't exist, job must be Failed
                # FIXME: it is not possible to return here an S_ERROR(), return the message only
                absentLFNs[lfn] = S_ERROR(errno.ENOENT, "File not in FC")["Message"]
        if absentLFNs:
            return S_OK(
                {
                    "onlineLFNs": list(onlineLFNs),
                    "offlineLFNs": offlineLFNsDict,
                    "failedLFNs": list(failedLFNs),
                    "absentLFNs": absentLFNs,
                }
            )
        return S_ERROR("Failures in getting replicas")

    lfnListReplicas = lfnListReplicas["Value"]["Successful"]
    # If a file is reported here at a tape SE, it is not at a disk SE as we use disk in priority
    # We shall check all file anyway in order to make sure they exist
    seToLFNs = dict()
    for lfn, ses in lfnListReplicas.items():
        for se in ses:
            seToLFNs.setdefault(se, list()).append(lfn)

    if seToLFNs:
        if jobState:
            # Get user name and group from the job state
            userName = jobState.getAttribute("Owner")
            if not userName["OK"]:
                return userName
            userName = userName["Value"]

            userGroup = jobState.getAttribute("OwnerGroup")
            if not userGroup["OK"]:
                return userGroup
            userGroup = userGroup["Value"]
        else:
            userName = None
            userGroup = None
        # Check whether files are Online or Offline, or missing at SE
        result = _checkFilesToStage(
            seToLFNs,
            onlineLFNs,
            offlineLFNs,
            absentLFNs,  # pylint: disable=unexpected-keyword-arg
            checkOnlyTapeSEs=checkOnlyTapeSEs,
            jobLog=jobLog,
            proxyUserName=userName,
            proxyUserGroup=userGroup,
            executionLock=True,
        )

        if not result["OK"]:
            return result
        failedLFNs = set(lfnList) - set(onlineLFNs) - set(offlineLFNs) - set(absentLFNs)

        # Get the online SEs
        dmsHelper = DMSHelpers()
        onlineSEs = set(se for ses in onlineLFNs.values() for se in ses)
        onlineSites = set(dmsHelper.getLocalSiteForSE(se).get("Value") for se in onlineSEs) - {None}
        for lfn in offlineLFNs:
            ses = offlineLFNs[lfn]
            if len(ses) == 1:
                # No choice, let's go
                offlineLFNsDict.setdefault(ses[0], list()).append(lfn)
                continue
            # Try and get an SE at a site already with online files
            found = False
            if onlineSites:
                # If there is at least one online site, select one
                for se in ses:
                    site = dmsHelper.getLocalSiteForSE(se)
                    if site["OK"]:
                        if site["Value"] in onlineSites:
                            offlineLFNsDict.setdefault(se, list()).append(lfn)
                            found = True
                            break
            # No online site found in common, select randomly
            if not found:
                offlineLFNsDict.setdefault(random.choice(ses), list()).append(lfn)

    return S_OK(
        {
            "onlineLFNs": list(onlineLFNs),
            "offlineLFNs": offlineLFNsDict,
            "failedLFNs": list(failedLFNs),
            "absentLFNs": absentLFNs,
            "onlineSites": onlineSites,
        }
    )


def _checkFilesToStage(
    seToLFNs,
    onlineLFNs,
    offlineLFNs,
    absentLFNs,
    checkOnlyTapeSEs=None,
    jobLog=None,
    proxyUserName=None,
    proxyUserGroup=None,
    executionLock=None,
):
    """
    Checks on SEs whether the file is NEARLINE or ONLINE
    onlineLFNs, offlineLFNs and absentLFNs are modified to contain the files found online
    If checkOnlyTapeSEs is True, disk replicas are not checked
    As soon as a replica is found Online for a file, no further check is made
    """
    # Only check on storage if it is a tape SE
    if jobLog is None:
        logger = gLogger
    else:
        logger = jobLog
    if checkOnlyTapeSEs is None:
        # Default value is True
        checkOnlyTapeSEs = True

    failed = {}
    for se, lfnsInSEList in seToLFNs.items():
        # If we have found already all files online at another SE, no need to check the others
        # but still we want to set the SE as Online if not a TapeSE
        vo = getVOForGroup(proxyUserGroup)
        seObj = StorageElement(se, vo=vo)
        status = seObj.getStatus()
        if not status["OK"]:
            return status
        tapeSE = status["Value"]["TapeSE"]
        diskSE = status["Value"]["DiskSE"]
        # If requested to check only Tape SEs and the file is at a diskSE, we guess it is Online...
        filesToCheck = []
        for lfn in lfnsInSEList:
            # If the file had already been found accessible at an SE, only check that this one is on disk
            diskIsOK = checkOnlyTapeSEs or (lfn in onlineLFNs)
            if diskIsOK and diskSE:
                onlineLFNs.setdefault(lfn, []).append(se)
            elif not diskIsOK or (tapeSE and (lfn not in onlineLFNs)):
                filesToCheck.append(lfn)
        if not filesToCheck:
            continue

        # We have to use a new SE object because it caches the proxy!
        with UserProxy(
            proxyUserName=proxyUserName, proxyUserGroup=proxyUserGroup, executionLock=executionLock
        ) as proxyResult:
            if proxyResult["OK"]:
                fileMetadata = StorageElement(se, vo=vo).getFileMetadata(filesToCheck)
            else:
                fileMetadata = proxyResult

        if not fileMetadata["OK"]:
            failed[se] = dict.fromkeys(filesToCheck, fileMetadata["Message"])
        else:
            if fileMetadata["Value"]["Failed"]:
                failed[se] = fileMetadata["Value"]["Failed"]
            # is there at least one replica online?
            for lfn, mDict in fileMetadata["Value"]["Successful"].items():
                # SRM returns Cached, but others may only return Accessible
                if mDict.get("Cached", mDict["Accessible"]):
                    onlineLFNs.setdefault(lfn, []).append(se)
                elif tapeSE:
                    # A file can be staged only at Tape SE
                    offlineLFNs.setdefault(lfn, []).append(se)
                else:
                    # File not available at a diskSE... we shall retry later
                    pass

    # Doesn't matter if some files are Offline if they are also online
    for lfn in set(offlineLFNs) & set(onlineLFNs):
        offlineLFNs.pop(lfn)

    # If the file was found staged, ignore possible errors, but print out errors
    for se, failedLfns in list(failed.items()):
        logger.error("Errors when getting files metadata", "at %s" % se)
        for lfn, reason in list(failedLfns.items()):
            if lfn in onlineLFNs:
                logger.warn(reason, "for %s, but there is an online replica" % lfn)
                failed[se].pop(lfn)
            else:
                logger.error(reason, "for %s, no online replicas" % lfn)
                if cmpError(reason, errno.ENOENT):
                    absentLFNs.setdefault(lfn, []).append(se)
                    failed[se].pop(lfn)
        if not failed[se]:
            failed.pop(se)
    # Find the files that do not exist at SE
    if failed:
        logger.error(
            "Error getting metadata", "for %d files" % len(set(lfn for lfnList in failed.values() for lfn in lfnList))
        )

    for lfn in absentLFNs:
        seList = absentLFNs[lfn]
        # FIXME: it is not possible to return here an S_ERROR(), return the message only
        absentLFNs[lfn] = S_ERROR(errno.ENOENT, "File not at %s" % ",".join(sorted(seList)))["Message"]
    # Format the error for absent files
    return S_OK()


@createClient("StorageManagement/StorageManager")
class StorageManagerClient(Client):
    """This is the client to the StorageManager service, so even if it is not seen, it exposes all its RPC calls"""

    def __init__(self, **kwargs):
        super(StorageManagerClient, self).__init__(**kwargs)
        self.setServer("StorageManagement/StorageManager")



""" FTS3Job module containing only the FTS3Job class """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

import datetime
import errno

# Requires at least version 3.3.3
import fts3.rest.client.easy as fts3
from fts3.rest.client.exceptions import FTS3ClientException, NotFound

# We specifically use Request in the FTS client because of a leak in the
# default pycurl. See https://its.cern.ch/jira/browse/FTS-261
from fts3.rest.client.request import Request as ftsSSLRequest

from DIRAC.Resources.Storage.StorageElement import StorageElement

from DIRAC.FrameworkSystem.Client.Logger import gLogger

from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.Utilities.DErrno import cmpError

from DIRAC.Core.Utilities.JEncode import JSerializable
from DIRAC.DataManagementSystem.Client.FTS3File import FTS3File

# 3 days in seconds
BRING_ONLINE_TIMEOUT = 259200


class FTS3Job(JSerializable):
    """Abstract class to represent a job to be executed by FTS. It belongs
    to an FTS3Operation
    """

    # START states

    # States from FTS doc https://fts3-docs.web.cern.ch/fts3-docs/docs/state_machine.html
    ALL_STATES = [
        "Submitted",  # Initial state of a job as soon it's dropped into the database
        "Ready",  # One of the files within a job went to Ready state
        "Active",  # One of the files within a job went to Active state
        "Finished",  # All files Finished gracefully
        "Canceled",  # Job canceled
        "Failed",  # All files Failed
        "Finisheddirty",  # Some files Failed
        "Staging",  # One of the files within a job went to Staging state
        "Archiving",  # From FTS: one of the files within a job went to Archiving state
    ]

    FINAL_STATES = ["Canceled", "Failed", "Finished", "Finisheddirty"]
    INIT_STATE = "Submitted"

    # END states
    _attrToSerialize = [
        "jobID",
        "operationID",
        "status",
        "error",
        "submitTime",
        "lastUpdate",
        "ftsServer",
        "ftsGUID",
        "completeness",
        "username",
        "userGroup",
    ]

    def __init__(self):

        self.submitTime = None
        self.lastUpdate = None
        self.lastMonitor = None

        self.ftsGUID = None
        self.ftsServer = None

        self.error = None
        self.status = FTS3Job.INIT_STATE

        self.completeness = None
        self.operationID = None

        self.username = None
        self.userGroup = None

        # temporary used only for submission
        # Set by FTS Operation when preparing
        self.type = None  # Transfer, Staging

        self.sourceSE = None
        self.targetSE = None
        self.filesToSubmit = []
        self.activity = None
        self.priority = None
        self.vo = None
        self.rmsReqID = None

        # temporary used only for accounting
        # it is set by the monitor method
        # when a job is in a final state
        self.accountingDict = None

    def monitor(self, context=None, ftsServer=None, ucert=None):
        """Queries the fts server to monitor the job.
        The internal state of the object is updated depending on the
        monitoring result.

        In case the job is not found on the server, the status is set to 'Failed'

        Within a job, only the transfers having a `fileID` metadata are considered.
        This is to allow for multihop jobs doing a staging

        This method assumes that the attribute self.ftsGUID is set

        :param context: fts3 context. If not given, it is created (see ftsServer & ucert param)
        :param ftsServer: the address of the fts server to submit to. Used only if context is
                          not given. if not given either, use the ftsServer object attribute

        :param ucert: path to the user certificate/proxy. Might be infered by the fts cli (see its doc)

        :returns: {FileID: { status, error } }

                  Possible error numbers

                  * errno.ESRCH: If the job does not exist on the server
                  * errno.EDEADLK: In case the job and file status are inconsistent (see comments inside the code)


        """

        if not self.ftsGUID:
            return S_ERROR("FTSGUID not set, FTS job not submitted?")

        if not context:
            if not ftsServer:
                ftsServer = self.ftsServer
            context = fts3.Context(endpoint=ftsServer, ucert=ucert, request_class=ftsSSLRequest, verify=False)

        jobStatusDict = None
        try:
            jobStatusDict = fts3.get_job_status(context, self.ftsGUID, list_files=True)
        # The job is not found
        # Set its status to Failed and return
        except NotFound:
            self.status = "Failed"
            return S_ERROR(errno.ESRCH, "FTSGUID %s not found on %s" % (self.ftsGUID, self.ftsServer))
        except FTS3ClientException as e:
            return S_ERROR("Error getting the job status %s" % e)

        now = datetime.datetime.utcnow().replace(microsecond=0)
        self.lastMonitor = now

        newStatus = jobStatusDict["job_state"].capitalize()
        if newStatus != self.status:
            self.status = newStatus
            self.lastUpdate = now
            self.error = jobStatusDict["reason"]

        if newStatus in self.FINAL_STATES:
            self._fillAccountingDict(jobStatusDict)

        filesInfoList = jobStatusDict["files"]
        filesStatus = {}
        statusSummary = {}

        # Make a copy, since we are potentially
        # deleting objects
        for fileDict in list(filesInfoList):
            file_state = fileDict["file_state"].capitalize()
            file_metadata = fileDict["file_metadata"]

            # previous version of the code did not have dictionary as
            # file_metadata
            if isinstance(file_metadata, dict):
                file_id = file_metadata.get("fileID")
            else:
                file_id = file_metadata

            # The transfer does not have a fileID attached to it
            # so it does not correspond to a file in our DB: skip it
            # (typical of jobs with different staging protocol == CTA)
            # We also remove it from the fileInfoList, such that it is
            # not considered for accounting
            if not file_id:
                filesInfoList.remove(fileDict)
                continue

            file_error = fileDict["reason"]
            filesStatus[file_id] = {"status": file_state, "error": file_error}

            # If the state of the file is final for FTS, set ftsGUID of the file to None,
            # such that it is "released" from this job and not updated anymore in future
            # monitoring calls
            if file_state in FTS3File.FTS_FINAL_STATES:
                filesStatus[file_id]["ftsGUID"] = None

            # If the file is not in a final state, but the job is, we return an error
            # FTS can have inconsistencies where the FTS Job is in a final state
            # but not all the files.
            # The inconsistencies are cleaned every hour on the FTS side.
            # https://its.cern.ch/jira/browse/FTS-1482
            elif self.status in self.FINAL_STATES:
                return S_ERROR(
                    errno.EDEADLK,
                    "Job %s in a final state (%s) while File %s is not (%s)"
                    % (self.ftsGUID, self.status, file_id, file_state),
                )

            statusSummary[file_state] = statusSummary.get(file_state, 0) + 1

        # We've removed all the intermediate transfers that we are not interested in
        # so we put this back into the monitoring data such that the accounting is done properly
        jobStatusDict["files"] = filesInfoList
        if newStatus in self.FINAL_STATES:
            self._fillAccountingDict(jobStatusDict)

        total = len(filesInfoList)
        completed = sum([statusSummary.get(state, 0) for state in FTS3File.FTS_FINAL_STATES])
        self.completeness = int(100 * completed / total)

        return S_OK(filesStatus)

    @staticmethod
    def __fetchSpaceToken(seName, vo):
        """Fetch the space token of storage element

        :param seName: name of the storageElement
        :param vo: vo of the job
        :returns: space token. If there is no SpaceToken defined, returns None
        """
        seToken = None
        if seName:
            seObj = StorageElement(seName, vo=vo)

            res = seObj.getStorageParameters(protocol="srm")
            if not res["OK"]:
                # If there is no SRM protocol, we do not specify
                # the space token
                if cmpError(res, errno.ENOPROTOOPT):
                    return S_OK(None)

                return res

            seToken = res["Value"].get("SpaceToken")

        return S_OK(seToken)

    @staticmethod
    def __isTapeSE(seName, vo):
        """Check whether a given SE is a tape storage

        :param seName: name of the storageElement
        :param vo: VO of the job
        :returns: True/False
                  In case of error, returns True.
                  It is better to lose a bit of time on the FTS
                  side, rather than failing jobs because the FTS default
                  pin time is too short
        """
        isTape = StorageElement(seName, vo=vo).getStatus().get("Value", {}).get("TapeSE", True)

        return isTape

    def _constructTransferJob(self, pinTime, allLFNs, target_spacetoken, protocols=None):
        """Build a job for transfer

        Some attributes of the job are expected to be set
          * sourceSE
          * targetSE
          * activity (optional)
          * priority (optional)
          * filesToSubmit
          * operationID (optional, used as metadata for the job)


        :param pinTime: pining time in case staging is needed
        :param allLFNs: list of LFNs to transfer
        :param failedLFNs: set of LFNs in filesToSubmit for which there was a problem
        :param target_spacetoken: the space token of the target
        :param protocols: list of protocols to restrict the protocol choice for the transfer

        :return: S_OK( (job object, list of ftsFileIDs in the job))
        """

        log = gLogger.getSubLogger(
            "constructTransferJob/%s/%s_%s" % (self.operationID, self.sourceSE, self.targetSE), True
        )

        res = self.__fetchSpaceToken(self.sourceSE, self.vo)
        if not res["OK"]:
            return res
        source_spacetoken = res["Value"]

        failedLFNs = set()
        dstSE = StorageElement(self.targetSE, vo=self.vo)
        srcSE = StorageElement(self.sourceSE, vo=self.vo)

        # If the source is not a tape SE, we should set the
        # copy_pin_lifetime and bring_online params to None,
        # otherwise they will do an extra useless queue in FTS
        sourceIsTape = self.__isTapeSE(self.sourceSE, self.vo)
        copy_pin_lifetime = pinTime if sourceIsTape else None
        bring_online = BRING_ONLINE_TIMEOUT if sourceIsTape else None
        archive_timeout = None

        # getting all the (source, dest) surls
        res = dstSE.generateTransferURLsBetweenSEs(allLFNs, srcSE, protocols=protocols)

        if not res["OK"]:
            return res

        for lfn, reason in res["Value"]["Failed"].items():
            failedLFNs.add(lfn)
            log.error("Could not get source SURL", "%s %s" % (lfn, reason))

        allSrcDstSURLs = res["Value"]["Successful"]
        srcProto, destProto = res["Value"]["Protocols"]

        # If the destination is a tape, and the protocol supports it,
        # check if we want to have an archive timeout
        dstIsTape = self.__isTapeSE(self.targetSE, self.vo)
        if dstIsTape and destProto in dstSE.localStageProtocolList:
            archive_timeout = dstSE.options.get("ArchiveTimeout")

        # This contains the staging URLs if they are different from the transfer URLs
        # (CTA...)
        allStageURLs = dict()

        # In case we are transfering from a tape system, and the stage protocol
        # is not the same as the transfer protocol, we generate the staging URLs
        # to do a multihop transfer. See below.
        if sourceIsTape and srcProto not in srcSE.localStageProtocolList:

            # As of version 3.10, FTS can only handle one file per multi hop
            # job. If we are here, that means that we need one, so make sure that
            # we only have a single file to transfer (this should have been checked
            # at the job construction step in FTS3Operation).
            # This test is important, because multiple files would result in the source
            # being deleted !
            if len(allLFNs) != 1:
                log.debug("Multihop job has %s files while only 1 allowed" % len(allLFNs))
                return S_ERROR(errno.E2BIG, "Trying multihop job with more than one file !")

            res = srcSE.getURL(allSrcDstSURLs, protocol=srcSE.localStageProtocolList)

            if not res["OK"]:
                return res

            for lfn, reason in res["Value"]["Failed"].items():
                failedLFNs.add(lfn)
                log.error("Could not get stage SURL", "%s %s" % (lfn, reason))
                allSrcDstSURLs.pop(lfn)

            allStageURLs = res["Value"]["Successful"]

        transfers = []

        fileIDsInTheJob = []

        for ftsFile in self.filesToSubmit:

            if ftsFile.lfn in failedLFNs:
                log.debug("Not preparing transfer for file %s" % ftsFile.lfn)
                continue

            sourceSURL, targetSURL = allSrcDstSURLs[ftsFile.lfn]
            stageURL = allStageURLs.get(ftsFile.lfn)

            if sourceSURL == targetSURL:
                log.error("sourceSURL equals to targetSURL", "%s" % ftsFile.lfn)
                ftsFile.error = "sourceSURL equals to targetSURL"
                ftsFile.status = "Defunct"
                continue

            ftsFileID = getattr(ftsFile, "fileID")

            # Under normal circumstances, we simply submit an fts transfer as such:
            # * srcProto://myFile -> destProto://myFile
            #
            # Even in case of the source storage being a tape system, it works fine.
            # However, if the staging and transfer protocols are different (which might be the case for CTA),
            #  we use the multihop machinery to submit two sequential fts transfers:
            # one to stage, one to transfer.
            # It looks like such
            # * stageProto://myFile -> stageProto://myFile
            # * srcProto://myFile -> destProto://myFile

            if stageURL:

                # We do not set a fileID in the metadata
                # such that we do not update the DB when monitoring
                stageTrans_metadata = {"desc": "PreStage %s" % ftsFileID}

                # If we use an activity, also set it as file metadata
                # for WLCG monitoring purposes
                # https://its.cern.ch/jira/projects/DOMATPC/issues/DOMATPC-14?
                if self.activity:
                    stageTrans_metadata["activity"] = self.activity

                stageTrans = fts3.new_transfer(
                    stageURL,
                    stageURL,
                    checksum="ADLER32:%s" % ftsFile.checksum,
                    filesize=ftsFile.size,
                    metadata=stageTrans_metadata,
                    activity=self.activity,
                )
                transfers.append(stageTrans)

            trans_metadata = {"desc": "Transfer %s" % ftsFileID, "fileID": ftsFileID}

            # If we use an activity, also set it as file metadata
            # for WLCG monitoring purposes
            # https://its.cern.ch/jira/projects/DOMATPC/issues/DOMATPC-14?
            if self.activity:
                trans_metadata["activity"] = self.activity

            # because of an xroot bug (https://github.com/xrootd/xrootd/issues/1433)
            # the checksum needs to be lowercase. It does not impact the other
            # protocol, so it's fine to put it here.
            # I only add it in this transfer and not the "staging" one above because it
            # impacts only root -> root transfers
            trans = fts3.new_transfer(
                sourceSURL,
                targetSURL,
                checksum="ADLER32:%s" % ftsFile.checksum.lower(),
                filesize=ftsFile.size,
                metadata=trans_metadata,
                activity=self.activity,
            )

            transfers.append(trans)
            fileIDsInTheJob.append(ftsFileID)

        if not transfers:
            log.error("No transfer possible!")
            return S_ERROR("No transfer possible")

        # We add a few metadata to the fts job so that we can reuse them later on without
        # querying our DB.
        # source and target SE are just used for accounting purpose
        job_metadata = {
            "operationID": self.operationID,
            "rmsReqID": self.rmsReqID,
            "sourceSE": self.sourceSE,
            "targetSE": self.targetSE,
        }

        if self.activity:
            job_metadata["activity"] = self.activity

        job = fts3.new_job(
            transfers=transfers,
            overwrite=True,
            source_spacetoken=source_spacetoken,
            spacetoken=target_spacetoken,
            bring_online=bring_online,
            copy_pin_lifetime=copy_pin_lifetime,
            retry=3,
            verify_checksum="target",  # Only check target vs specified, since we verify the source earlier
            multihop=bool(allStageURLs),  # if we have stage urls, then we need multihop
            metadata=job_metadata,
            priority=self.priority,
            archive_timeout=archive_timeout,
        )

        return S_OK((job, fileIDsInTheJob))

    # def _constructRemovalJob(self, context, allLFNs, failedLFNs, target_spacetoken):
    #   """ Build a job for removal
    #
    #       Some attributes of the job are expected to be set
    #         * targetSE
    #         * activity (optional)
    #         * priority (optional)
    #         * filesToSubmit
    #         * operationID (optional, used as metadata for the job)
    #
    #
    #       :param context: fts3 context
    #       :param allLFNs: List of LFNs to remove
    #       :param failedLFNs: set of LFNs in filesToSubmit for which there was a problem
    #       :param target_spacetoken: the space token of the target
    #
    #       :return: S_OK( (job object, list of ftsFileIDs in the job))
    #   """
    #
    #   log = gLogger.getSubLogger(
    #       "constructRemovalJob/%s/%s" %
    #       (self.operationID, self.targetSE), True)
    #
    #   transfers = []
    #   fileIDsInTheJob = []
    #   for ftsFile in self.filesToSubmit:
    #
    #     if ftsFile.lfn in failedLFNs:
    #       log.debug("Not preparing transfer for file %s" % ftsFile.lfn)
    #       continue
    #
    #     transfers.append({'surl': allTargetSURLs[ftsFile.lfn],
    #                       'metadata': getattr(ftsFile, 'fileID')})
    #     fileIDsInTheJob.append(getattr(ftsFile, 'fileID'))
    #
    #   # We add a few metadata to the fts job so that we can reuse them later on without
    #   # querying our DB.
    #   # source and target SE are just used for accounting purpose
    #   job_metadata = {
    #       'operationID': self.operationID,
    #       'sourceSE': self.sourceSE,
    #       'targetSE': self.targetSE}
    #
    #   job = fts3.new_delete_job(transfers,
    #                             spacetoken=target_spacetoken,
    #                             metadata=job_metadata)
    #   job['params']['retry'] = 3
    #   job['params']['priority'] = self.priority
    #
    #   return S_OK((job, fileIDsInTheJob))

    def _constructStagingJob(self, pinTime, allLFNs, target_spacetoken):
        """Build a job for staging

        Some attributes of the job are expected to be set
          * targetSE
          * activity (optional)
          * priority (optional)
          * filesToSubmit
          * operationID (optional, used as metadata for the job)

        :param pinTime: pining time in case staging is needed
        :param allLFNs: List of LFNs to stage
        :param failedLFNs: set of LFNs in filesToSubmit for which there was a problem
        :param target_spacetoken: the space token of the target

        :return: S_OK( (job object, list of ftsFileIDs in the job))
        """

        log = gLogger.getSubLogger("constructStagingJob/%s/%s" % (self.operationID, self.targetSE), True)

        transfers = []
        fileIDsInTheJob = []

        # Set of LFNs for which we did not get an SRM URL
        failedLFNs = set()

        # getting all the target surls
        res = StorageElement(self.targetSE, vo=self.vo).getURL(allLFNs, protocol="srm")
        if not res["OK"]:
            return res

        for lfn, reason in res["Value"]["Failed"].items():
            failedLFNs.add(lfn)
            log.error("Could not get target SURL", "%s %s" % (lfn, reason))

        allTargetSURLs = res["Value"]["Successful"]

        for ftsFile in self.filesToSubmit:

            if ftsFile.lfn in failedLFNs:
                log.debug("Not preparing transfer for file %s" % ftsFile.lfn)
                continue

            sourceSURL = targetSURL = allTargetSURLs[ftsFile.lfn]
            ftsFileID = getattr(ftsFile, "fileID")
            trans_metadata = {"desc": "Stage %s" % ftsFileID, "fileID": ftsFileID}
            trans = fts3.new_transfer(
                sourceSURL,
                targetSURL,
                checksum="ADLER32:%s" % ftsFile.checksum,
                filesize=ftsFile.size,
                metadata=trans_metadata,
                activity=self.activity,
            )

            transfers.append(trans)
            fileIDsInTheJob.append(ftsFileID)

        # If the source is not an tape SE, we should set the
        # copy_pin_lifetime and bring_online params to None,
        # otherwise they will do an extra useless queue in FTS
        sourceIsTape = self.__isTapeSE(self.sourceSE, self.vo)
        copy_pin_lifetime = pinTime if sourceIsTape else None
        bring_online = 86400 if sourceIsTape else None

        # We add a few metadata to the fts job so that we can reuse them later on without
        # querying our DB.
        # source and target SE are just used for accounting purpose
        job_metadata = {"operationID": self.operationID, "sourceSE": self.sourceSE, "targetSE": self.targetSE}

        if self.activity:
            job_metadata["activity"] = self.activity

        job = fts3.new_job(
            transfers=transfers,
            overwrite=True,
            source_spacetoken=target_spacetoken,
            spacetoken=target_spacetoken,
            bring_online=bring_online,
            copy_pin_lifetime=copy_pin_lifetime,
            retry=3,
            metadata=job_metadata,
            priority=self.priority,
        )

        return S_OK((job, fileIDsInTheJob))

    def submit(self, context=None, ftsServer=None, ucert=None, pinTime=36000, protocols=None):
        """submit the job to the FTS server

        Some attributes are expected to be defined for the submission to work:
          * type (set by FTS3Operation)
          * sourceSE (only for Transfer jobs)
          * targetSE
          * activity (optional)
          * priority (optional)
          * username
          * userGroup
          * filesToSubmit
          * operationID (optional, used as metadata for the job)

        We also expect the FTSFiles have an ID defined, as it is given as transfer metadata

        :param pinTime: Time the file should be pinned on disk (used for transfers and staging)
                        Used only if he source SE is a tape storage
        :param context: fts3 context. If not given, it is created (see ftsServer & ucert param)
        :param ftsServer: the address of the fts server to submit to. Used only if context is
                          not given. if not given either, use the ftsServer object attribute

        :param ucert: path to the user certificate/proxy. Might be inferred by the fts cli (see its doc)
        :param protocols: list of protocols from which we should choose the protocol to use

        :returns: S_OK([FTSFiles ids of files submitted])
        """

        log = gLogger.getLocalSubLogger("submit/%s/%s_%s" % (self.operationID, self.sourceSE, self.targetSE))

        if not context:
            if not ftsServer:
                ftsServer = self.ftsServer
            context = fts3.Context(endpoint=ftsServer, ucert=ucert, request_class=ftsSSLRequest, verify=False)

        # Construct the target SURL
        res = self.__fetchSpaceToken(self.targetSE, self.vo)
        if not res["OK"]:
            return res
        target_spacetoken = res["Value"]

        allLFNs = [ftsFile.lfn for ftsFile in self.filesToSubmit]

        if self.type == "Transfer":
            res = self._constructTransferJob(pinTime, allLFNs, target_spacetoken, protocols=protocols)
        elif self.type == "Staging":
            res = self._constructStagingJob(pinTime, allLFNs, target_spacetoken)
        # elif self.type == 'Removal':
        #   res = self._constructRemovalJob(context, allLFNs, failedLFNs, target_spacetoken)

        if not res["OK"]:
            return res

        job, fileIDsInTheJob = res["Value"]
        setFileIdsInTheJob = set(fileIDsInTheJob)

        try:
            self.ftsGUID = fts3.submit(context, job)
            log.info("Got GUID %s" % self.ftsGUID)

            # Only increase the amount of attempt
            # if we succeeded in submitting -> no ! Why did I do that ??
            for ftsFile in self.filesToSubmit:
                ftsFile.attempt += 1

                # This should never happen because a file should be "released"
                # first by the previous job.
                # But we just print a warning
                if ftsFile.ftsGUID is not None:
                    log.warn(
                        "FTSFile has a non NULL ftsGUID at job submission time",
                        "FileID: %s existing ftsGUID: %s" % (ftsFile.fileID, ftsFile.ftsGUID),
                    )

                # `assign` the file to this job
                ftsFile.ftsGUID = self.ftsGUID
                if ftsFile.fileID in setFileIdsInTheJob:
                    ftsFile.status = "Submitted"

            now = datetime.datetime.utcnow().replace(microsecond=0)
            self.submitTime = now
            self.lastUpdate = now
            self.lastMonitor = now

        except FTS3ClientException as e:
            log.exception("Error at submission", repr(e))
            return S_ERROR("Error at submission: %s" % e)

        return S_OK(fileIDsInTheJob)

    @staticmethod
    def generateContext(ftsServer, ucert, lifetime=25200):
        """This method generates an fts3 context

        :param ftsServer: address of the fts3 server
        :param ucert: the path to the certificate to be used
        :param lifetime: duration (in sec) of the delegation to the FTS3 server
                        (default is 7h, like FTS3 default)

        :returns: an fts3 context
        """
        try:
            context = fts3.Context(endpoint=ftsServer, ucert=ucert, request_class=ftsSSLRequest, verify=False)

            # Explicitely delegate to be sure we have the lifetime we want
            # Note: the delegation will re-happen only when the FTS server
            # decides that there is not enough timeleft.
            # At the moment, this is 1 hour, which effectively means that if you do
            # not submit a job for more than 1h, you have no valid proxy in FTS servers
            # anymore. In future release of FTS3, the delegation will be triggered when
            # one third of the lifetime will be left.
            # Also, the proxy given as parameter might have less than "lifetime" left
            # since it is cached, but it does not matter, because in the FTS3Agent
            # we make sure that we renew it often enough
            # Finally, FTS3 has an issue with handling the lifetime of the proxy,
            # because it does not check all the chain. This is under discussion
            # https://its.cern.ch/jira/browse/FTS-1575
            fts3.delegate(context, lifetime=datetime.timedelta(seconds=lifetime))

            return S_OK(context)
        except FTS3ClientException as e:
            gLogger.exception("Error generating context", repr(e))
            return S_ERROR(repr(e))

    def _fillAccountingDict(self, jobStatusDict):
        """This methods generates the necessary information to create a DataOperation
        accounting record, and stores them as a instance attribute.

        For it to be relevant, it should be called only when the job is in a final state.

        :param jobStatusDict: output of fts3.get_job_status

        :returns: None
        """

        accountingDict = dict()
        sourceSE = None
        targetSE = None

        accountingDict["OperationType"] = "ReplicateAndRegister"

        accountingDict["User"] = self.username
        accountingDict["Protocol"] = "FTS3"
        accountingDict["ExecutionSite"] = self.ftsServer

        # We cannot rely on all the transient attributes (like self.filesToSubmit)
        # because it is probably not filed by the time we monitor !

        filesInfoList = jobStatusDict["files"]
        successfulFiles = []

        for fileDict in filesInfoList:
            file_state = fileDict["file_state"].capitalize()
            if file_state in FTS3File.FTS_SUCCESS_STATES:
                successfulFiles.append(fileDict)

        job_metadata = jobStatusDict["job_metadata"]
        # previous version of the code did not have dictionary as
        # job_metadata
        if isinstance(job_metadata, dict):
            sourceSE = job_metadata.get("sourceSE")
            targetSE = job_metadata.get("targetSE")

        accountingDict["TransferOK"] = len(successfulFiles)
        accountingDict["TransferTotal"] = len(filesInfoList)
        # We need this if in the list comprehension because staging only jobs have `None` as filesize
        accountingDict["TransferSize"] = sum(
            [fileDict["filesize"] for fileDict in successfulFiles if fileDict["filesize"]]
        )
        accountingDict["FinalStatus"] = self.status
        accountingDict["Source"] = sourceSE
        accountingDict["Destination"] = targetSE
        # We need this if in the list comprehension because staging only jobs have `None` as tx_duration
        accountingDict["TransferTime"] = sum(
            int(fileDict["tx_duration"]) for fileDict in successfulFiles if fileDict["tx_duration"]
        )

        # Registration values must be set anyway
        accountingDict["RegistrationTime"] = 0.0
        accountingDict["RegistrationOK"] = 0
        accountingDict["RegistrationTotal"] = 0

        self.accountingDict = accountingDict



########################################################################
# File  : Watchdog.py
# Author: Stuart Paterson
########################################################################

"""  The Watchdog class is used by the Job Wrapper to resolve and monitor
     the system resource consumption.  The Watchdog can determine if
     a running job is stalled and indicate this to the Job Wrapper.
     Furthermore, the Watchdog will identify when the Job CPU limit has been
     exceeded and fail jobs meaningfully.

     Information is returned to the WMS via the heart-beat mechanism.  This
     also interprets control signals from the WMS e.g. to kill a running
     job.

     - Still to implement:
          - CPU normalization for correct comparison with job limit
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import re
import time
import resource
import errno
import socket
import getpass
import psutil

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities import MJF
from DIRAC.Core.Utilities.Profiler import Profiler
from DIRAC.Core.Utilities.Os import getDiskSpace
from DIRAC.Core.Utilities.Subprocess import getChildrenPIDs
from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getSystemInstance
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import TimeLeft
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient
from DIRAC.WorkloadManagementSystem.Client import JobMinorStatus


class Watchdog(object):

    #############################################################################
    def __init__(self, pid, exeThread, spObject, jobCPUTime, memoryLimit=0, processors=1, jobArgs={}):
        """Constructor, takes system flag as argument."""
        self.stopSigStartSeconds = int(jobArgs.get("StopSigStartSeconds", 1800))  # 30 minutes
        self.stopSigFinishSeconds = int(jobArgs.get("StopSigFinishSeconds", 1800))  # 30 minutes
        self.stopSigNumber = int(jobArgs.get("StopSigNumber", 2))  # SIGINT
        self.stopSigRegex = jobArgs.get("StopSigRegex", None)
        self.stopSigSent = False

        self.log = gLogger.getSubLogger("Watchdog")
        self.exeThread = exeThread
        self.wrapperPID = pid
        self.appPID = self.exeThread.getCurrentPID()
        self.spObject = spObject
        self.jobCPUTime = jobCPUTime
        self.memoryLimit = memoryLimit
        self.calibration = 0
        self.initialValues = {}
        self.parameters = {}
        self.peekFailCount = 0
        self.peekRetry = 5
        self.profiler = Profiler(pid)
        self.checkError = ""
        self.currentStats = {}
        self.initialized = False
        self.count = 0

        # defaults
        self.testWallClock = 1
        self.testDiskSpace = 1
        self.testLoadAvg = 1
        self.maxWallClockTime = 3 * 24 * 60 * 60
        self.testCPUConsumed = 1
        self.testCPULimit = 0
        self.testMemoryLimit = 0
        self.testTimeLeft = 1
        self.pollingTime = 10  # 10 seconds
        self.checkingTime = 30 * 60  # 30 minute period
        self.minCheckingTime = 20 * 60  # 20 mins
        self.wallClockCheckSeconds = 5 * 60  # 5 minutes
        self.maxWallClockTime = 3 * 24 * 60 * 60  # e.g. 4 days
        self.jobPeekFlag = 1  # on / off
        self.minDiskSpace = 10  # MB
        self.loadAvgLimit = 1000  # > 1000 and jobs killed
        self.sampleCPUTime = 30 * 60  # e.g. up to 20mins sample
        self.jobCPUMargin = 20  # %age buffer before killing job
        self.minCPUWallClockRatio = 5  # ratio %age
        self.nullCPULimit = 5  # After 5 sample times return null CPU consumption kill job
        self.checkCount = 0
        self.wallClockCheckCount = 0
        self.nullCPUCount = 0

        self.grossTimeLeftLimit = 10 * self.checkingTime
        self.timeLeftUtil = TimeLeft()
        self.timeLeft = 0
        self.littleTimeLeft = False
        self.cpuPower = 1.0
        self.processors = processors

    #############################################################################
    def initialize(self):
        """Watchdog initialization."""
        if self.initialized:
            self.log.info("Watchdog already initialized")
            return S_OK()
        else:
            self.initialized = True

        setup = gConfig.getValue("/DIRAC/Setup", "")
        if not setup:
            return S_ERROR("Can not get the DIRAC Setup value")
        wms_instance = getSystemInstance("WorkloadManagement")
        if not wms_instance:
            return S_ERROR("Can not get the WorkloadManagement system instance")
        self.section = "/Systems/WorkloadManagement/%s/JobWrapper" % wms_instance

        self.log.verbose("Watchdog initialization")
        # Test control flags
        self.testWallClock = gConfig.getValue(self.section + "/CheckWallClockFlag", 1)
        self.testDiskSpace = gConfig.getValue(self.section + "/CheckDiskSpaceFlag", 1)
        self.testLoadAvg = gConfig.getValue(self.section + "/CheckLoadAvgFlag", 1)
        self.testCPUConsumed = gConfig.getValue(self.section + "/CheckCPUConsumedFlag", 1)
        self.testCPULimit = gConfig.getValue(self.section + "/CheckCPULimitFlag", 0)
        self.testMemoryLimit = gConfig.getValue(self.section + "/CheckMemoryLimitFlag", 0)
        self.testTimeLeft = gConfig.getValue(self.section + "/CheckTimeLeftFlag", 1)
        # Other parameters
        self.pollingTime = gConfig.getValue(self.section + "/PollingTime", 10)  # 10 seconds
        self.checkingTime = gConfig.getValue(self.section + "/CheckingTime", 30 * 60)  # 30 minute period
        self.minCheckingTime = gConfig.getValue(self.section + "/MinCheckingTime", 20 * 60)  # 20 mins
        self.maxWallClockTime = gConfig.getValue(self.section + "/MaxWallClockTime", 3 * 24 * 60 * 60)  # e.g. 4 days
        self.jobPeekFlag = gConfig.getValue(self.section + "/JobPeekFlag", 1)  # on / off
        self.minDiskSpace = gConfig.getValue(self.section + "/MinDiskSpace", 10)  # MB
        self.loadAvgLimit = gConfig.getValue(self.section + "/LoadAverageLimit", 1000)  # > 1000 and jobs killed
        self.sampleCPUTime = gConfig.getValue(self.section + "/CPUSampleTime", 30 * 60)  # e.g. up to 20mins sample
        self.jobCPUMargin = gConfig.getValue(self.section + "/JobCPULimitMargin", 20)  # %age buffer before killing job
        self.minCPUWallClockRatio = gConfig.getValue(self.section + "/MinCPUWallClockRatio", 5)  # ratio %age
        # After 5 sample times return null CPU consumption kill job
        self.nullCPULimit = gConfig.getValue(self.section + "/NullCPUCountLimit", 5)
        if self.checkingTime < self.minCheckingTime:
            self.log.info(
                "Requested CheckingTime of %s setting to %s seconds (minimum)"
                % (self.checkingTime, self.minCheckingTime)
            )
            self.checkingTime = self.minCheckingTime

        # The time left is returned in seconds @ 250 SI00 = 1 HS06,
        # the self.checkingTime and self.pollingTime are in seconds,
        # thus they need to be multiplied by a large enough factor
        self.fineTimeLeftLimit = gConfig.getValue(self.section + "/TimeLeftLimit", 150 * self.pollingTime)
        self.cpuPower = gConfig.getValue("/LocalSite/CPUNormalizationFactor", 1.0)

        return S_OK()

    def run(self):
        """The main watchdog execution method"""

        result = self.initialize()
        if not result["OK"]:
            self.log.always("Can not start watchdog for the following reason")
            self.log.always(result["Message"])
            return result

        try:
            while True:
                self.log.debug("Starting watchdog loop # %d" % self.count)
                start_cycle_time = time.time()
                result = self.execute()
                exec_cycle_time = time.time() - start_cycle_time
                if not result["OK"]:
                    self.log.error("Watchdog error during execution", result["Message"])
                    break
                elif result["Value"] == "Ended":
                    break
                self.count += 1
                if exec_cycle_time < self.pollingTime:
                    time.sleep(self.pollingTime - exec_cycle_time)
            return S_OK()
        except Exception:
            self.log.exception()
            return S_ERROR("Exception")

    #############################################################################
    def execute(self):
        """The main agent execution method of the Watchdog."""

        if not self.exeThread.is_alive():
            self.__getUsageSummary()
            self.log.info("Process to monitor has completed, Watchdog will exit.")
            return S_OK("Ended")

        # WallClock checks every self.wallClockCheckSeconds, but only if StopSigRegex is defined in JDL
        if (
            not self.stopSigSent
            and self.stopSigRegex is not None
            and (time.time() - self.initialValues["StartTime"]) > self.wallClockCheckSeconds * self.wallClockCheckCount
        ):
            self.wallClockCheckCount += 1
            self._performWallClockChecks()

        if self.littleTimeLeft:
            # if we have gone over enough iterations query again
            if self.littleTimeLeftCount == 0 and self.__timeLeft() == -1:
                self.checkError = JobMinorStatus.JOB_EXCEEDED_CPU
                self.log.error(self.checkError, self.timeLeft)
                self.__killRunningThread()
                return S_OK()
            else:
                self.littleTimeLeftCount -= 1

        # Note: need to poll regularly to see if the thread is alive
        #      but only perform checks with a certain frequency
        if (time.time() - self.initialValues["StartTime"]) > self.checkingTime * self.checkCount:
            self.checkCount += 1
            result = self._performChecks()
            if not result["OK"]:
                self.log.warn("Problem during recent checks")
                self.log.warn(result["Message"])
            return S_OK()
        else:
            # self.log.debug('Application thread is alive: checking count is %s' %(self.checkCount))
            return S_OK()

    #############################################################################
    def _performWallClockChecks(self):
        """Watchdog performs the wall clock checks based on MJF. Signals are sent
        to processes if we need to stop, but function always returns S_OK()
        """
        mjf = MJF.MJF()

        try:
            wallClockSecondsLeft = mjf.getWallClockSecondsLeft()
        except Exception as e:
            # Just stop if we can't get the wall clock seconds left
            return S_OK()

        jobstartSeconds = mjf.getIntJobFeature("jobstart_secs")
        if jobstartSeconds is None:
            # Just stop if we don't know when the job started
            return S_OK()

        if (int(time.time()) > jobstartSeconds + self.stopSigStartSeconds) and (
            wallClockSecondsLeft < self.stopSigFinishSeconds + self.wallClockCheckSeconds
        ):
            # Need to send the signal! Assume it works to avoid sending the signal more than once
            self.log.info("Sending signal to JobWrapper children", "(%s)" % self.stopSigNumber)
            self.stopSigSent = True

            try:
                for childPid in getChildrenPIDs(self.wrapperPID):
                    try:
                        cmdline = open("/proc/%d/cmdline" % childPid, "r").read().replace("\0", " ").strip()
                    except IOError:
                        # Process gone away? Not running on Linux? Skip anyway
                        continue

                    if re.search(self.stopSigRegex, cmdline) is not None:
                        self.log.info(
                            'Sending signal %d to process ID %d, cmdline = "%s"'
                            % (self.stopSigNumber, childPid, cmdline)
                        )
                        os.kill(childPid, self.stopSigNumber)

            except Exception as e:
                self.log.error("Failed to send signals to JobWrapper children!", repr(e))

        return S_OK()

    #############################################################################
    def _performChecks(self):
        """The Watchdog checks are performed at a different period to the checking of the
        application thread and correspond to the checkingTime.
        """
        self.log.verbose("------------------------------------")
        self.log.verbose("Checking loop starts for Watchdog")
        heartBeatDict = {}

        msg = ""

        loadAvg = float(os.getloadavg()[0])
        msg += "LoadAvg: %d " % loadAvg
        heartBeatDict["LoadAverage"] = loadAvg
        if "LoadAverage" not in self.parameters:
            self.parameters["LoadAverage"] = []
        self.parameters["LoadAverage"].append(loadAvg)

        memoryUsed = self.getMemoryUsed()
        msg += "MemUsed: %.1f kb " % (memoryUsed)
        heartBeatDict["MemoryUsed"] = memoryUsed
        if "MemoryUsed" not in self.parameters:
            self.parameters["MemoryUsed"] = []
        self.parameters["MemoryUsed"].append(memoryUsed)

        result = self.profiler.vSizeUsage(withChildren=True)
        if not result["OK"]:
            self.log.warn("Could not get vSize info from profiler", result["Message"])
        else:
            vsize = result["Value"] * 1024.0
            heartBeatDict["Vsize"] = vsize
            self.parameters.setdefault("Vsize", [])
            self.parameters["Vsize"].append(vsize)
            msg += "Job Vsize: %.1f kb " % vsize

        result = self.profiler.memoryUsage(withChildren=True)
        if not result["OK"]:
            self.log.warn("Could not get rss info from profiler", result["Message"])
        else:
            rss = result["Value"] * 1024.0
            heartBeatDict["RSS"] = rss
            self.parameters.setdefault("RSS", [])
            self.parameters["RSS"].append(rss)
            msg += "Job RSS: %.1f kb " % rss

        if "DiskSpace" not in self.parameters:
            self.parameters["DiskSpace"] = []

        # We exclude fuse so that mountpoints can be cleaned up by automount after a period unused
        # (specific request from CERN batch service).
        result = self.getDiskSpace(exclude="fuse")
        if not result["OK"]:
            self.log.warn("Could not establish DiskSpace", result["Message"])
        else:
            msg += "DiskSpace: %.1f MB " % (result["Value"])
            self.parameters["DiskSpace"].append(result["Value"])
            heartBeatDict["AvailableDiskSpace"] = result["Value"]

        cpu = self.__getCPU()
        if not cpu["OK"]:
            msg += "CPU: ERROR "
            hmsCPU = 0
        else:
            cpu = cpu["Value"]
            msg += "CPU: %s (h:m:s) " % (cpu)
            if "CPUConsumed" not in self.parameters:
                self.parameters["CPUConsumed"] = []
            self.parameters["CPUConsumed"].append(cpu)
            hmsCPU = cpu
            rawCPU = self.__convertCPUTime(hmsCPU)
            if rawCPU["OK"]:
                heartBeatDict["CPUConsumed"] = rawCPU["Value"]

        result = self.__getWallClockTime()
        if not result["OK"]:
            self.log.warn("Failed determining wall clock time", result["Message"])
        else:
            msg += "WallClock: %.2f s " % (result["Value"])
            self.parameters.setdefault("WallClockTime", list()).append(result["Value"])
            heartBeatDict["WallClockTime"] = result["Value"] * self.processors
        self.log.info(msg)

        result = self._checkProgress()
        if not result["OK"]:
            self.checkError = result["Message"]
            self.log.warn(self.checkError)

            if self.jobPeekFlag:
                result = self.__peek()
                if result["OK"]:
                    outputList = result["Value"]
                    self.log.info("Last lines of available application output:")
                    self.log.info("================START================")
                    for line in outputList:
                        self.log.info(line)

                    self.log.info("=================END=================")

            self.__killRunningThread()
            return S_OK()

        recentStdOut = "None"
        if self.jobPeekFlag:
            result = self.__peek()
            if result["OK"]:
                outputList = result["Value"]
                size = len(outputList)
                recentStdOut = "Last %s lines of application output from Watchdog on %s [UTC]:" % (
                    size,
                    Time.dateTime(),
                )
                border = "=" * len(recentStdOut)
                cpuTotal = "Last reported CPU consumed for job is %s (h:m:s)" % (hmsCPU)
                if self.timeLeft:
                    cpuTotal += ", Batch Queue Time Left %s (s @ HS06)" % self.timeLeft
                recentStdOut = "\n%s\n%s\n%s\n%s\n" % (border, recentStdOut, cpuTotal, border)
                self.log.info(recentStdOut)
                for line in outputList:
                    self.log.info(line)
                    recentStdOut += line + "\n"
            else:
                recentStdOut = (
                    "Watchdog is initializing and will attempt to obtain standard output from application thread"
                )
                self.log.info(recentStdOut)
                self.peekFailCount += 1
                if self.peekFailCount > self.peekRetry:
                    self.jobPeekFlag = 0
                    self.log.warn("Turning off job peeking for remainder of execution")

        if "JOBID" not in os.environ:
            self.log.info("Running without JOBID so parameters will not be reported")
            return S_OK()

        jobID = os.environ["JOBID"]
        staticParamDict = {"StandardOutput": recentStdOut}
        self.__sendSignOfLife(int(jobID), heartBeatDict, staticParamDict)
        return S_OK("Watchdog checking cycle complete")

    #############################################################################
    def __getCPU(self):
        """Uses the profiler to get CPU time for current process, its child, and the terminated child,
        and returns HH:MM:SS after conversion.
        """
        result = self.profiler.cpuUsageUser(withChildren=True, withTerminatedChildren=True)
        if not result["OK"]:
            self.log.warn("Issue while checking consumed CPU for user", result["Message"])
            if result["Errno"] == errno.ESRCH:
                self.log.warn("The main process does not exist (anymore). This might be correct.")
            return result
        cpuUsageUser = result["Value"]

        result = self.profiler.cpuUsageSystem(withChildren=True, withTerminatedChildren=True)
        if not result["OK"]:
            self.log.warn("Issue while checking consumed CPU for system", result["Message"])
            if result["Errno"] == errno.ESRCH:
                self.log.warn("The main process does not exist (anymore). This might be correct.")
            return result
        cpuUsageSystem = result["Value"]

        cpuTimeTotal = cpuUsageUser + cpuUsageSystem
        if cpuTimeTotal:
            self.log.verbose("Raw CPU time consumed (s) =", cpuTimeTotal)
            return self.__getCPUHMS(cpuTimeTotal)
        self.log.error("CPU time consumed found to be 0")
        return S_ERROR()

    #############################################################################
    def __getCPUHMS(self, cpuTime):
        mins, secs = divmod(cpuTime, 60)
        hours, mins = divmod(mins, 60)
        humanTime = "%02d:%02d:%02d" % (hours, mins, secs)
        self.log.verbose("Human readable CPU time is: %s" % humanTime)
        return S_OK(humanTime)

    #############################################################################
    def __interpretControlSignal(self, signalDict):
        """This method is called whenever a signal is sent via the result of
        sending a sign of life.
        """
        self.log.info("Received control signal")
        if isinstance(signalDict, dict):
            if "Kill" in signalDict:
                self.log.info("Received Kill signal, stopping job via control signal")
                self.checkError = JobMinorStatus.RECEIVED_KILL_SIGNAL
                self.__killRunningThread()
            else:
                self.log.info("The following control signal was sent but not understood by the watchdog:")
                self.log.info(signalDict)
        else:
            self.log.info("Expected dictionary for control signal", "received:\n%s" % (signalDict))

        return S_OK()

    #############################################################################
    def _checkProgress(self):
        """This method calls specific tests to determine whether the job execution
        is proceeding normally.  CS flags can easily be added to add or remove
        tests via central configuration.
        """
        report = ""

        if self.testWallClock:
            result = self.__checkWallClockTime()
            if not result["OK"]:
                self.log.warn(result["Message"])
                return result
            report += "WallClock: OK, "
        else:
            report += "WallClock: NA,"

        if self.testDiskSpace:
            result = self.__checkDiskSpace()
            if not result["OK"]:
                self.log.warn(result["Message"])
                return result
            report += "DiskSpace: OK, "
        else:
            report += "DiskSpace: NA,"

        if self.testLoadAvg:
            result = self.__checkLoadAverage()
            if not result["OK"]:
                self.log.warn(
                    "Check of load average failed, but won't fail because of that", ": %s" % result["Message"]
                )
                report += "LoadAverage: ERROR, "
                return S_OK()
            report += "LoadAverage: OK, "
        else:
            report += "LoadAverage: NA,"

        if self.testCPUConsumed:
            result = self.__checkCPUConsumed()
            if not result["OK"]:
                return result
            report += "CPUConsumed: OK, "
        else:
            report += "CPUConsumed: NA, "

        if self.testCPULimit:
            result = self.__checkCPULimit()
            if not result["OK"]:
                self.log.warn(result["Message"])
                return result
            report += "CPULimit OK, "
        else:
            report += "CPULimit: NA, "

        if self.testTimeLeft:
            self.__timeLeft()
            if self.timeLeft:
                report += "TimeLeft: OK"
        else:
            report += "TimeLeft: NA"

        if self.testMemoryLimit:
            result = self.__checkMemoryLimit()
            if not result["OK"]:
                self.log.warn(result["Message"])
                return result
            report += "MemoryLimit OK, "
        else:
            report += "MemoryLimit: NA, "

        self.log.info(report)
        return S_OK("All enabled checks passed")

    #############################################################################
    def __checkCPUConsumed(self):
        """Checks whether the CPU consumed by application process is reasonable. This
        method will report stalled jobs to be killed.
        """
        self.log.info("Checking CPU Consumed")

        if "WallClockTime" not in self.parameters:
            return S_ERROR("Missing WallClockTime info")
        if "CPUConsumed" not in self.parameters:
            return S_ERROR("Missing CPUConsumed info")

        wallClockTime = self.parameters["WallClockTime"][-1]
        if wallClockTime < self.sampleCPUTime:
            self.log.info(
                "Stopping check, wallclock time is still smaller than sample time",
                "(%s) < (%s)" % (wallClockTime, self.sampleCPUTime),
            )
            return S_OK()

        intervals = max(1, int(self.sampleCPUTime / self.checkingTime))
        if len(self.parameters["CPUConsumed"]) < intervals + 1:
            self.log.info(
                "Not enough snapshots to calculate",
                "there are %s and we need %s" % (len(self.parameters["CPUConsumed"]), intervals + 1),
            )
            return S_OK()

        wallClockTime = self.parameters["WallClockTime"][-1] - self.parameters["WallClockTime"][-1 - intervals]
        try:
            cpuTime = self.__convertCPUTime(self.parameters["CPUConsumed"][-1])["Value"]
            # For some reason, some times the CPU consumed estimation returns 0
            # if cpuTime == 0:
            #   return S_OK()
            cpuTime -= self.__convertCPUTime(self.parameters["CPUConsumed"][-1 - intervals])["Value"]
            if cpuTime < 0:
                self.log.warn("Consumed CPU time negative, something wrong may have happened, ignore")
                return S_OK()
            if wallClockTime <= 0:
                self.log.warn("Wallclock time should not be negative or zero, Ignore")
                return S_OK()

            ratio = (cpuTime / wallClockTime) * 100

            self.log.info("CPU/Wallclock ratio is %.2f%%" % ratio)
            # in case of error cpuTime might be 0, exclude this
            if ratio < self.minCPUWallClockRatio:
                if (
                    os.path.exists("DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK")
                    or "DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK" in os.environ
                ):
                    self.log.warn(
                        "N.B. job would be declared as stalled but CPU / WallClock check is disabled by payload"
                    )
                    return S_OK()
                self.log.info("Job is stalled!")
                return S_ERROR(JobMinorStatus.WATCHDOG_STALLED)
        except Exception as e:
            self.log.error("Cannot convert CPU consumed from string to int", str(e))

        return S_OK()

    #############################################################################

    def __convertCPUTime(self, cputime):
        """Method to convert the CPU time as returned from the Watchdog
        instances to the equivalent DIRAC normalized CPU time to be compared
        to the Job CPU requirement.
        """
        cpuValue = 0
        cpuHMS = cputime.split(":")
        # for i in range( len( cpuHMS ) ):
        #   cpuHMS[i] = cpuHMS[i].replace( '00', '0' )

        try:
            hours = float(cpuHMS[0]) * 60 * 60
            mins = float(cpuHMS[1]) * 60
            secs = float(cpuHMS[2])
            cpuValue = float(hours + mins + secs)
        except Exception as x:
            self.log.warn(str(x))
            return S_ERROR("Could not calculate CPU time")

        # Normalization to be implemented
        normalizedCPUValue = cpuValue

        result = S_OK()
        result["Value"] = normalizedCPUValue
        self.log.debug("CPU value %s converted to %s" % (cputime, normalizedCPUValue))
        return result

    #############################################################################

    def __checkCPULimit(self):
        """Checks that the job has consumed more than the job CPU requirement
        (plus a configurable margin) and kills them as necessary.
        """
        consumedCPU = 0
        if "CPUConsumed" in self.parameters:
            consumedCPU = self.parameters["CPUConsumed"][-1]

        consumedCPUDict = self.__convertCPUTime(consumedCPU)
        if consumedCPUDict["OK"]:
            currentCPU = consumedCPUDict["Value"]
        else:
            return S_OK("Not possible to determine current CPU consumed")

        if consumedCPU:
            limit = int(self.jobCPUTime + self.jobCPUTime * (self.jobCPUMargin / 100))
            cpuConsumed = float(currentCPU)
            if cpuConsumed > limit:
                self.log.info(
                    "Job has consumed more than the specified CPU limit",
                    "with an additional %s%% margin" % (self.jobCPUMargin),
                )
                return S_ERROR("Job has exceeded maximum CPU time limit")
            return S_OK("Job within CPU limit")

        if not currentCPU:
            self.log.verbose("Both initial and current CPU consumed are null")
            return S_OK("CPU consumed is not measurable yet")

        return S_OK("Not possible to determine CPU consumed")

    def __checkMemoryLimit(self):
        """Checks that the job memory consumption is within a limit"""
        vsize = 0
        if "Vsize" in self.parameters:
            vsize = self.parameters["Vsize"][-1]

        if vsize and self.memoryLimit:
            if vsize > self.memoryLimit:
                vsize = vsize
                # Just a warning for the moment
                self.log.warn(
                    "Job has consumed %f.2 KB of memory with the limit of %f.2 KB" % (vsize, self.memoryLimit)
                )

        return S_OK()

    #############################################################################
    def __checkDiskSpace(self):
        """Checks whether the CS defined minimum disk space is available."""
        if "DiskSpace" in self.parameters:
            availSpace = self.parameters["DiskSpace"][-1]
            if availSpace >= 0 and availSpace < self.minDiskSpace:
                self.log.info(
                    "Not enough local disk space for job to continue, defined in CS as %s MB" % (self.minDiskSpace)
                )
                return S_ERROR(JobMinorStatus.JOB_INSUFFICIENT_DISK)
            else:
                return S_OK("Job has enough disk space available")
        else:
            return S_ERROR("Available disk space could not be established")

    #############################################################################
    def __checkWallClockTime(self):
        """Checks whether the job has been running for the CS defined maximum
        wall clock time.
        """
        if "StartTime" in self.initialValues:
            startTime = self.initialValues["StartTime"]
            if time.time() - startTime > self.maxWallClockTime:
                self.log.info("Job has exceeded maximum wall clock time of %s seconds" % (self.maxWallClockTime))
                return S_ERROR(JobMinorStatus.JOB_EXCEEDED_WALL_CLOCK)
            else:
                return S_OK("Job within maximum wall clock time")
        else:
            return S_ERROR("Job start time could not be established")

    #############################################################################
    def __checkLoadAverage(self):
        """Checks whether the CS defined maximum load average is exceeded."""
        if "LoadAverage" in self.parameters:
            loadAvg = self.parameters["LoadAverage"][-1]
            if loadAvg > float(self.loadAvgLimit):
                self.log.info("Maximum load average exceeded, defined in CS as %s " % (self.loadAvgLimit))
                return S_ERROR("Job exceeded maximum load average")
            return S_OK("Job running with normal load average")
        return S_ERROR("Job load average not established")

    #############################################################################
    def __peek(self):
        """Uses ExecutionThread.getOutput() method to obtain standard output
        from running thread via subprocess callback function.
        """
        result = self.exeThread.getOutput()
        if not result["OK"]:
            self.log.warn("Could not obtain output from running application thread")
            self.log.warn(result["Message"])

        return result

    #############################################################################
    def calibrate(self):
        """The calibrate method obtains the initial values for system memory and load
        and calculates the margin for error for the rest of the Watchdog cycle.
        """
        self.__getWallClockTime()
        self.parameters["WallClockTime"] = []

        cpuConsumed = self.__getCPU()
        if not cpuConsumed["OK"]:
            self.log.warn("Could not establish CPU consumed, setting to 0.0")
            cpuConsumed = 0.0
        else:
            cpuConsumed = cpuConsumed["Value"]

        self.initialValues["CPUConsumed"] = cpuConsumed
        self.parameters["CPUConsumed"] = []

        self.initialValues["LoadAverage"] = float(os.getloadavg()[0])
        self.parameters["LoadAverage"] = []

        memUsed = self.getMemoryUsed()

        self.initialValues["MemoryUsed"] = memUsed
        self.parameters["MemoryUsed"] = []

        result = self.profiler.vSizeUsage(withChildren=True)
        if not result["OK"]:
            self.log.warn("Could not get vSize info from profiler", result["Message"])
        else:
            vsize = result["Value"] * 1024.0
            self.initialValues["Vsize"] = vsize
            self.log.verbose("Vsize(kb)", "%.1f" % vsize)
        self.parameters["Vsize"] = []

        result = self.profiler.memoryUsage(withChildren=True)
        if not result["OK"]:
            self.log.warn("Could not get rss info from profiler", result["Message"])
        else:
            rss = result["Value"] * 1024.0
            self.initialValues["RSS"] = rss
            self.log.verbose("RSS(kb)", "%.1f" % rss)
        self.parameters["RSS"] = []

        # We exclude fuse so that mountpoints can be cleaned up by automount after a period unused
        # (specific request from CERN batch service).
        result = self.getDiskSpace(exclude="fuse")
        self.log.verbose("DiskSpace: %s" % (result))
        if not result["OK"]:
            self.log.warn("Could not establish DiskSpace")
        else:
            self.initialValues["DiskSpace"] = result["Value"]
        self.parameters["DiskSpace"] = []

        result = self.getNodeInformation()
        self.log.verbose("NodeInfo", result)

        if "LSB_JOBID" in os.environ:
            result["LocalJobID"] = os.environ["LSB_JOBID"]
        if "PBS_JOBID" in os.environ:
            result["LocalJobID"] = os.environ["PBS_JOBID"]
        if "QSUB_REQNAME" in os.environ:
            result["LocalJobID"] = os.environ["QSUB_REQNAME"]
        if "JOB_ID" in os.environ:
            result["LocalJobID"] = os.environ["JOB_ID"]

        self.__reportParameters(result, "NodeInformation", True)
        self.__reportParameters(self.initialValues, "InitialValues")
        return S_OK()

    def __timeLeft(self):
        """
        return Normalized CPU time left in the batch system
        0 if not available
        update self.timeLeft and self.littleTimeLeft
        """
        # Get CPU time left in the batch system
        result = self.timeLeftUtil.getTimeLeft(0.0)
        if not result["OK"]:
            # Could not get CPU time left, we might need to wait for the first loop
            # or the Utility is not working properly for this batch system
            # or we are in a batch system
            timeLeft = 0
        else:
            timeLeft = result["Value"]

        self.timeLeft = timeLeft
        if not self.littleTimeLeft:
            if timeLeft and timeLeft < self.grossTimeLeftLimit:
                self.log.info("TimeLeft bellow %s, now checking with higher frequency" % timeLeft)
                self.littleTimeLeft = True
                # TODO: better configurable way of doing this to be coded
                self.littleTimeLeftCount = 15
        else:
            if self.timeLeft and self.timeLeft < self.fineTimeLeftLimit:
                timeLeft = -1

        return timeLeft

    #############################################################################
    def __getUsageSummary(self):
        """Returns average load, memory etc. over execution of job thread"""
        summary = {}
        # CPUConsumed
        if "CPUConsumed" in self.parameters:
            cpuList = self.parameters["CPUConsumed"]
            if cpuList:
                hmsCPU = cpuList[-1]
                rawCPU = self.__convertCPUTime(hmsCPU)
                if rawCPU["OK"]:
                    summary["LastUpdateCPU(s)"] = rawCPU["Value"]
            else:
                summary["LastUpdateCPU(s)"] = "Could not be estimated"
        # DiskSpace
        if "DiskSpace" in self.parameters:
            space = self.parameters["DiskSpace"]
            if space:
                value = abs(float(space[-1]) - float(self.initialValues["DiskSpace"]))
                if value < 0:
                    value = 0
                summary["DiskSpace(MB)"] = value
            else:
                summary["DiskSpace(MB)"] = "Could not be estimated"
        # MemoryUsed
        if "MemoryUsed" in self.parameters:
            memory = self.parameters["MemoryUsed"]
            if memory:
                summary["MemoryUsed(kb)"] = abs(float(memory[-1]) - float(self.initialValues["MemoryUsed"]))
            else:
                summary["MemoryUsed(kb)"] = "Could not be estimated"
        # LoadAverage
        if "LoadAverage" in self.parameters:
            laList = self.parameters["LoadAverage"]
            if laList:
                summary["LoadAverage"] = sum(laList) / len(laList)
            else:
                summary["LoadAverage"] = "Could not be estimated"

        result = self.__getWallClockTime()
        if not result["OK"]:
            self.log.warn("Failed determining wall clock time", result["Message"])
            summary["WallClockTime(s)"] = 0
            summary["ScaledCPUTime(s)"] = 0
        else:
            wallClock = result["Value"]
            summary["WallClockTime(s)"] = wallClock * self.processors
            summary["ScaledCPUTime(s)"] = wallClock * self.cpuPower * self.processors

        self.__reportParameters(summary, "UsageSummary", True)
        self.currentStats = summary

    #############################################################################
    def __reportParameters(self, params, title=None, report=False):
        """Will report parameters for job."""
        try:
            parameters = []
            self.log.info("", "==========================================================")
            if title:
                self.log.info("Watchdog will report", title)
            else:
                self.log.info("Watchdog will report parameters")
            self.log.info("", "==========================================================")
            vals = params
            if "Value" in params:
                if vals["Value"]:
                    vals = params["Value"]
            for k, v in vals.items():
                if v:
                    self.log.info(str(k) + " = " + str(v))
                    parameters.append([k, v])
            if report:
                self.__setJobParamList(parameters)

            self.log.info("", "==========================================================")
        except Exception as x:
            self.log.warn("Problem while reporting parameters")
            self.log.warn(repr(x))

    #############################################################################
    def __getWallClockTime(self):
        """Establishes the Wall Clock time spent since the Watchdog initialization"""
        result = S_OK()
        if "StartTime" in self.initialValues:
            currentTime = time.time()
            wallClock = currentTime - self.initialValues["StartTime"]
            result["Value"] = wallClock
        else:
            self.initialValues["StartTime"] = time.time()
            result["Value"] = 0.0

        return result

    #############################################################################
    def __killRunningThread(self):
        """Will kill the running thread process and any child processes."""
        self.log.info("Sending kill signal to application PID %s" % (self.spObject.getChildPID()))
        result = self.spObject.killChild()
        self.applicationKilled = True
        self.log.info("Subprocess.killChild() returned:%s " % (result))
        return S_OK("Thread killed")

    #############################################################################
    def __sendSignOfLife(self, jobID, heartBeatDict, staticParamDict):
        """Sends sign of life 'heartbeat' signal and triggers control signal
        interpretation.
        """
        result = JobStateUpdateClient().sendHeartBeat(jobID, heartBeatDict, staticParamDict)
        if not result["OK"]:
            self.log.warn("Problem sending sign of life")
            self.log.warn(result)

        if result["OK"] and result["Value"]:
            self.__interpretControlSignal(result["Value"])

        return result

    #############################################################################
    def __setJobParamList(self, value):
        """Wraps around setJobParameters of state update client"""
        # job wrapper template sets the jobID variable
        if "JOBID" not in os.environ:
            self.log.info("Running without JOBID so parameters will not be reported")
            return S_OK()
        jobID = os.environ["JOBID"]
        jobParam = JobStateUpdateClient().setJobParameters(int(jobID), value)
        self.log.verbose("setJobParameters(%s,%s)" % (jobID, value))
        if not jobParam["OK"]:
            self.log.warn(jobParam["Message"])

        return jobParam

    #############################################################################
    def getNodeInformation(self):
        """Retrieves all static system information"""
        result = {}
        result["HostName"] = socket.gethostname()
        result["CPU(MHz)"] = psutil.cpu_freq()[0]
        result["Memory(kB)"] = int(psutil.virtual_memory()[1] / 1024)
        result["LocalAccount"] = getpass.getuser()

        with open("/proc/cpuinfo", "r") as cpuinfo:
            info = cpuinfo.readlines()
        result["ModelName"] = info[4].split(":")[1].replace(" ", "").replace("\n", "")
        result["CacheSize(kB)"] = [x.strip().split(":")[1] for x in info if "cache size" in x][0].strip()

        return result

    #############################################################################
    def getMemoryUsed(self):
        """Obtains the memory used."""
        mem = (
            resource.getrusage(resource.RUSAGE_SELF).ru_maxrss + resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss
        )
        return float(mem)

    #############################################################################
    def getDiskSpace(self, exclude=None):
        """Obtains the available disk space."""
        result = S_OK()
        diskSpace = getDiskSpace(exclude=exclude)

        if diskSpace == -1:
            result = S_ERROR("Could not obtain disk usage")
            self.log.warn(" Could not obtain disk usage")
            result["Value"] = float(-1)
            return result

        result["Value"] = float(diskSpace)
        return result


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



#!/usr/bin/env python
"""
Get computing resources capable to execute a job with the given description.

Note that only statically defined computing resource parameters are considered although sites
can fail matching due to their dynamic state, e.g. occupancy by other jobs. Also input data
proximity is not taken into account.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, gLogger, exit as DIRACExit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

__RCSID__ = "$Id$"

fullMatch = False
sites = None


def setFullMatch(optVal_):
    global fullMatch
    fullMatch = True
    return S_OK()


def setSites(optVal_):
    global sites
    sites = optVal_.split(",")
    return S_OK()


@Script()
def main():
    global fullMatch
    global sites
    Script.registerSwitch("F", "full-match", "Check all the matching criteria", setFullMatch)
    Script.registerSwitch("S:", "site=", "Check matching for these sites (comma separated list)", setSites)
    Script.registerArgument("job_JDL: file with job JDL description")
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup
    from DIRAC.ConfigurationSystem.Client.Helpers import Resources
    from DIRAC.Core.Utilities.PrettyPrint import printTable
    from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
    from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus
    from DIRAC.WorkloadManagementSystem.Utilities.QueueUtilities import getQueuesResolved, matchQueue

    with open(args[0]) as f:
        jdl = f.read()

    # Get the current VO
    result = getVOfromProxyGroup()
    if not result["OK"]:
        gLogger.error("No proxy found, please login")
        DIRACExit(-1)
    voName = result["Value"]

    resultQueues = Resources.getQueues(siteList=sites, community=voName)
    if not resultQueues["OK"]:
        gLogger.error("Failed to get CE information")
        DIRACExit(-1)
    siteDict = resultQueues["Value"]
    result = getQueuesResolved(siteDict, {}, checkPlatform=True)
    if not resultQueues["OK"]:
        gLogger.error("Failed to get CE information")
        DIRACExit(-1)
    queueDict = result["Value"]

    # get list of usable sites within this cycle
    resultMask = SiteStatus().getUsableSites()
    if not resultMask["OK"]:
        gLogger.error("Failed to get Site mask information")
        DIRACExit(-1)
    siteMaskList = resultMask.get("Value", [])

    rssClient = ResourceStatus()

    fields = ("Site", "CE", "Queue", "Status", "Match", "Reason")
    records = []

    for queue, queueInfo in queueDict.items():
        site = queueInfo["Site"]
        ce = queueInfo["CEName"]
        siteStatus = "Active" if site in siteMaskList else "InActive"
        ceStatus = siteStatus
        if rssClient.rssFlag:
            result = rssClient.getElementStatus(ce, "ComputingElement")
            if result["OK"]:
                ceStatus = result["Value"][ce]["all"]

        result = matchQueue(jdl, queueInfo["ParametersDict"], fullMatch=fullMatch)
        if not result["OK"]:
            gLogger.error("Failed in getting match data", result["Message"])
            DIRACExit(-1)
        status = "Active" if siteStatus == "Active" and ceStatus == "Active" else "Inactive"
        if result["Value"]["Match"]:
            records.append((site, ce, queueInfo["QueueName"], status, "Yes", ""))
        else:
            records.append((site, ce, queueInfo["QueueName"], status, "No", result["Value"]["Reason"]))

    gLogger.notice(printTable(fields, records, sortField="Site", columnSeparator="  ", printOut=False))


if __name__ == "__main__":
    main()



"""Collection of DIRAC useful list related modules.
   By default on Error they return None.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import random
import sys


def uniqueElements(aList):
    """Utility to retrieve list of unique elements in a list (order is kept).

    :param aList: list of elements
    :type aList: python:list
    :return: list of unique elements
    """
    result = []
    seen = set()
    try:
        for i in aList:
            if i not in seen:
                result.append(i)
                seen.add(i)
        return result
    except Exception:
        return None


def appendUnique(aList, anObject):
    """Append to list if object does not exist.

    :param aList: list of elements
    :type aList: python:list
    :param anObject: object you want to append
    """
    if anObject not in aList:
        aList.append(anObject)


def fromChar(inputString, sepChar=","):
    """Generates a list splitting a string by the required character(s)
    resulting string items are stripped and empty items are removed.

    :param string inputString: list serialised to string
    :param string sepChar: separator
    :return: list of strings or None if sepChar has a wrong type
    """
    # to prevent getting an empty String as argument
    if not (isinstance(inputString, six.string_types) and isinstance(sepChar, six.string_types) and sepChar):
        return None
    return [fieldString.strip() for fieldString in inputString.split(sepChar) if len(fieldString.strip()) > 0]


def randomize(aList):
    """Return a randomly sorted list.

    :param aList: list to permute
    :type aList: python:list
    """
    tmpList = list(aList)
    random.shuffle(tmpList)
    return tmpList


def pop(aList, popElement):
    """Pop the first element equal to popElement from the list.

    :param aList: list
    :type aList: python:list
    :param popElement: element to pop
    """
    if popElement in aList:
        return aList.pop(aList.index(popElement))


def stringListToString(aList):
    """This function is used for making MySQL queries with a list of string elements.

    :param aList: list to be serialized to string for making queries
    :type aList: python:list
    """
    return ",".join("'%s'" % x for x in aList)


def intListToString(aList):
    """This function is used for making MySQL queries with a list of int elements.

    :param aList: list to be serialized to string for making queries
    :type aList: python:list
    """
    return ",".join(str(x) for x in aList)


def getChunk(aList, chunkSize):
    """Generator yielding chunk from a list of a size chunkSize.

    :param aList: list to be splitted
    :type aList: python:list
    :param int chunkSize: lenght of one chunk
    :raise: StopIteration

    Usage:

    >>> for chunk in getChunk( aList, chunkSize=10):
          process( chunk )

    """
    chunkSize = int(chunkSize)
    for i in range(0, len(aList), chunkSize):
        yield aList[i : i + chunkSize]


def breakListIntoChunks(aList, chunkSize):
    """This function takes a list as input and breaks it into list of size 'chunkSize'.
       It returns a list of lists.

    :param aList: list of elements
    :type aList: python:list
    :param int chunkSize: len of a single chunk
    :return: list of lists of length of chunkSize
    :raise: RuntimeError if numberOfFilesInChunk is less than 1
    """
    if chunkSize < 1:
        raise RuntimeError("chunkSize cannot be less than 1")
    if isinstance(aList, (set, dict, tuple, {}.keys().__class__, {}.items().__class__, {}.values().__class__)):
        aList = list(aList)
    return [chunk for chunk in getChunk(aList, chunkSize)]


def getIndexInList(anItem, aList):
    """Return the index of the element x in the list l
    or sys.maxint if it does not exist

    :param anItem: element to look for
    :param list aList: list to look into

    :return: the index or sys.maxint
    """
    # try:
    if anItem in aList:
        return aList.index(anItem)
    else:
        return sys.maxsize
    # except ValueError:
    #   return sys.maxsize



""" DataStore is the service for inserting accounting reports (rows) in the Accounting DB

    This service CAN be duplicated iff the first is a "master" and all the others are slaves.
    See the information about :ref:`datastorehelpers`.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN DataStore
  :end-before: ##END
  :dedent: 2
  :caption: DataStore options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import datetime
import six

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.AccountingSystem.DB.MultiAccountingDB import MultiAccountingDB
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Core.DISET.RPCClient import RPCClient

__RCSID__ = "$Id$"


class DataStoreHandler(RequestHandler):
    """DISET implementation of service for inserting records in accountingDB."""

    __acDB = None

    @classmethod
    def initializeHandler(cls, svcInfoDict):
        multiPath = PathFinder.getDatabaseSection("Accounting/MultiDB")
        cls.__acDB = MultiAccountingDB(multiPath)
        # we can run multiple services in read only mode. In that case we do not bucket
        cls.runBucketing = getServiceOption(svcInfoDict, "RunBucketing", True)
        if cls.runBucketing:
            cls.__acDB.autoCompactDB()  # pylint: disable=no-member
            result = cls.__acDB.markAllPendingRecordsAsNotTaken()  # pylint: disable=no-member
            if not result["OK"]:
                return result
            gThreadScheduler.addPeriodicTask(60, cls.__acDB.loadPendingRecords)  # pylint: disable=no-member
        return S_OK()

    types_registerType = [six.string_types, list, list, list]

    def export_registerType(self, typeName, definitionKeyFields, definitionAccountingFields, bucketsLength):
        """
        Register a new type. (Only for all powerful admins)
        (Bow before me for I am admin! :)
        """
        retVal = gConfig.getSections("/DIRAC/Setups")
        if not retVal["OK"]:
            return retVal
        errorsList = []
        for setup in retVal["Value"]:
            retVal = self.__acDB.registerType(  # pylint: disable=no-member
                setup, typeName, definitionKeyFields, definitionAccountingFields, bucketsLength
            )
            if not retVal["OK"]:
                errorsList.append(retVal["Message"])
        if errorsList:
            return S_ERROR("Error while registering type:\n %s" % "\n ".join(errorsList))
        return S_OK()

    types_setBucketsLength = [six.string_types, list]

    def export_setBucketsLength(self, typeName, bucketsLength):
        """
        Change the buckets Length. (Only for all powerful admins)
        (Bow before me for I am admin! :)
        """
        retVal = gConfig.getSections("/DIRAC/Setups")
        if not retVal["OK"]:
            return retVal
        errorsList = []
        for setup in retVal["Value"]:
            retVal = self.__acDB.changeBucketsLength(setup, typeName, bucketsLength)  # pylint: disable=no-member
            if not retVal["OK"]:
                errorsList.append(retVal["Message"])
        if errorsList:
            return S_ERROR("Error while changing bucketsLength type:\n %s" % "\n ".join(errorsList))
        return S_OK()

    types_regenerateBuckets = [six.string_types]

    def export_regenerateBuckets(self, typeName):
        """
        Recalculate buckets. (Only for all powerful admins)
        (Bow before me for I am admin! :)
        """
        retVal = gConfig.getSections("/DIRAC/Setups")
        if not retVal["OK"]:
            return retVal
        errorsList = []
        for setup in retVal["Value"]:
            retVal = self.__acDB.regenerateBuckets(setup, typeName)  # pylint: disable=no-member
            if not retVal["OK"]:
                errorsList.append(retVal["Message"])
        if errorsList:
            return S_ERROR("Error while recalculating buckets for type:\n %s" % "\n ".join(errorsList))
        return S_OK()

    types_getRegisteredTypes = []

    def export_getRegisteredTypes(self):
        """
        Get a list of registered types (Only for all powerful admins)
        (Bow before me for I am admin! :)
        """
        return self.__acDB.getRegisteredTypes()  # pylint: disable=no-member

    types_deleteType = [six.string_types]

    def export_deleteType(self, typeName):
        """
        Delete accounting type and ALL its contents. VERY DANGEROUS! (Only for all powerful admins)
        (Bow before me for I am admin! :)
        """
        retVal = gConfig.getSections("/DIRAC/Setups")
        if not retVal["OK"]:
            return retVal
        errorsList = []
        for setup in retVal["Value"]:
            retVal = self.__acDB.deleteType(setup, typeName)  # pylint: disable=too-many-function-args,no-member
            if not retVal["OK"]:
                errorsList.append(retVal["Message"])
        if errorsList:
            return S_ERROR("Error while deleting type:\n %s" % "\n ".join(errorsList))
        return S_OK()

    types_commit = [six.string_types, datetime.datetime, datetime.datetime, list]

    def export_commit(self, typeName, startTime, endTime, valuesList):
        """
        Add a record for a type
        """
        setup = self.serviceInfoDict["clientSetup"]
        startTime = int(Time.toEpoch(startTime))
        endTime = int(Time.toEpoch(endTime))
        return self.__acDB.insertRecordThroughQueue(  # pylint: disable=no-member
            setup, typeName, startTime, endTime, valuesList
        )

    types_commitRegisters = [list]

    def export_commitRegisters(self, entriesList):
        """
        Add a record for a type
        """
        setup = self.serviceInfoDict["clientSetup"]
        expectedTypes = [six.string_types, datetime.datetime, datetime.datetime, list]
        for entry in entriesList:
            if len(entry) != 4:
                return S_ERROR("Invalid records")
            for i, _ in enumerate(entry):
                if not isinstance(entry[i], expectedTypes[i]):
                    self.log.error(
                        "Unexpected type in report",
                        ": field %d in the records should be %s (and it is %s)" % (i, expectedTypes[i], type(entry[i])),
                    )
                    return S_ERROR("Unexpected type in report")
        records = []
        for entry in entriesList:
            startTime = int(Time.toEpoch(entry[1]))
            endTime = int(Time.toEpoch(entry[2]))
            self.log.debug("inserting", entry)
            records.append((setup, entry[0], startTime, endTime, entry[3]))
        return self.__acDB.insertRecordBundleThroughQueue(records)

    types_compactDB = []

    def export_compactDB(self):
        """
        Compact the db by grouping buckets
        """
        # if we are running slaves (not only one service) we can redirect the request to the master
        # For more information please read the Administrative guide Accounting part!
        # ADVICE: If you want to trigger the bucketing, please make sure the bucketing is not running!!!!
        if self.runBucketing:
            return self.__acDB.compactBuckets()  # pylint: disable=no-member

        return RPCClient("Accounting/DataStoreMaster").compactDB()

    types_remove = [six.string_types, datetime.datetime, datetime.datetime, list]

    def export_remove(self, typeName, startTime, endTime, valuesList):
        """
        Remove a record for a type
        """
        setup = self.serviceInfoDict["clientSetup"]
        startTime = int(Time.toEpoch(startTime))
        endTime = int(Time.toEpoch(endTime))
        return self.__acDB.deleteRecord(setup, typeName, startTime, endTime, valuesList)  # pylint: disable=no-member

    types_removeRegisters = [list]

    def export_removeRegisters(self, entriesList):
        """
        Remove a record for a type
        """
        setup = self.serviceInfoDict["clientSetup"]
        expectedTypes = [six.string_types, datetime.datetime, datetime.datetime, list]
        for entry in entriesList:
            if len(entry) != 4:
                return S_ERROR("Invalid records")
            for i in range(len(entry)):
                if not isinstance(entry[i], expectedTypes[i]):
                    return S_ERROR("%s field in the records should be %s" % (i, expectedTypes[i]))
        ok = 0
        for entry in entriesList:
            startTime = int(Time.toEpoch(entry[1]))
            endTime = int(Time.toEpoch(entry[2]))
            record = entry[3]
            result = self.__acDB.deleteRecord(setup, entry[0], startTime, endTime, record)  # pylint: disable=no-member
            if not result["OK"]:
                return S_OK(ok)
            ok += 1

        return S_OK(ok)
